#!/usr/bin/env python

import os
import copy
import time
import logging
import pandas as pd
from tqdm import tqdm
import multiprocessing
import traceback
from collections import defaultdict
import numpy as np
import scipy.spatial.distance

import inStrain
import inStrain.logUtils
import inStrain.readComparer

def run_compare_multiprocessing(SC_queue, result_queue, null_model, num_to_run, **kwargs):
    """
    Run the multiprocessing associated with the "compare" operation
    """
    # Get kwargs
    p = int(kwargs.get('processes', 6))

    # Set up progress bar
    results = []
    pbar = tqdm(desc='Comparing scaffolds: ', total=num_to_run)

    # Launch the multiprocessing
    processes = []
    inStrain.logUtils.log_checkpoint("Compare", "multiprocessing", "start")
    if p > 1:
        ctx = multiprocessing.get_context('spawn')
        for i in range(0, p):
            processes.append(ctx.Process(target=run_SC_objects, args=(SC_queue, result_queue, null_model, kwargs)))
        for proc in processes:
            proc.start()
    else:
        run_SC_objects(SC_queue, result_queue, null_model, kwargs, single_thread=True)

    # Get the results
    recieved_groups = 0
    while recieved_groups < num_to_run:
        result, log = result_queue.get()
        logging.debug(log)
        if result is not None:
            results.append(result)
        recieved_groups += 1
        pbar.update(1)

    # Close multi-processing
    for proc in processes:
        proc.terminate()

    # Close progress bar
    pbar.close()
    inStrain.logUtils.log_checkpoint("Compare", "multiprocessing", "end")

    return results

def run_SC_objects(cmd_queue, result_queue, null_model, kwargs, single_thread=False):
    """
    The actually paralellized unit for running compare
    """
    # Apply multiprocessing patch
    inStrain.controller.patch_mp_connection_bpo_17560()

    # Continually push as you run
    while True:
        # Get an SC object
        if single_thread:
            try:
                cmd = cmd_queue.get(timeout=5)
            except:
                return
        else:
            cmd = cmd_queue.get(True)

        # Process that SC object
        SC_result, log = SC_object_wrapper(cmd, null_model, kwargs)
        result_queue.put((SC_result, log))

        # Clean up memory
        for r in SC_result:
            del r

def SC_object_wrapper(SC, null_model, kwargs):
    '''
    Take a Scaffold Compare object and profile the scaffold
    '''
    debug = kwargs.get('debug', False)
    try:
        results, log = inStrain.readComparer.compare_scaffold(SC.scaffold, SC.names, SC.SNPtables, SC.covTs,
                                                                      SC.length, null_model, **kwargs)

    except Exception as e:
        if debug:
            print(e)
            traceback.print_exc()
        logging.error("whole scaffold exception- {0}".format(str(SC.scaffold)))
        t = time.strftime('%m-%d %H:%M')
        log_message = "\n{1} DEBUG FAILURE CompareScaffold {0} {2}\n" \
            .format(SC.scaffold, t, str(SC.names))
        results = None
        log = log_message

    return (results, log)


def subset_SNP_table(db, scaffold):
    if len(db) > 0:
        db = db[db['scaffold'] == scaffold]
        if len(db) == 0:
            db = pd.DataFrame()
        else:
            db = db.sort_values('mm')
            #db = db[['position', 'mm', 'con_base', 'ref_base', 'var_base', 'position_coverage', 'A', 'C', 'T', 'G', 'allele_count']]
    else:
        db = pd.DataFrame()

    return db

def find_relevant_scaffolds(input, bts, kwargs):
    """
    Return a list of scaffolds in the input based on the parameters of kwargs
    """
    GIdb = inStrain.SNVprofile.SNVprofile(input).get('genome_level_info')
    if GIdb is None:
        logging.error(f"Profile {input} does not have genome-level information; needed to run compare in database mode")
        raise Exception
    if 'mm' in GIdb:
        GIdb = GIdb.sort_values('mm', ascending=True).drop_duplicates(subset=['genome'], keep='last')

    min_breadth = kwargs.get('breadth', 0.5)
    genomes = GIdb[GIdb['breadth_minCov'] >= min_breadth]['genome'].tolist()

    scaffolds = []
    for genome in genomes:
        if genome in bts:
            scaffolds += bts[genome]
        else:
            logging.error(f'{genome} is in input {input} but not the provided stb file!')
            raise Exception(f'{genome} is in input {input} but not the provided stb file!')

    message = f'{input} has {len(genomes)} genomes detected and {len(scaffolds)} scaffolds'
    print(message)
    logging.info(message)

    return set(scaffolds)

def add_av_RC(db, v1='popANI', v2='coverage_overlap', n1='av_ani', n2='av_cov'):
    '''
    add a column titled 'av_ani' to the passed in dataframe
    dataframe must have rows reference, querey, and ani
    Args:
        db: dataframe
    '''

    combo2value = defaultdict(lambda: np.nan)
    combo2value2 = defaultdict(lambda: np.nan)
    for i, row in db.iterrows():
        combo2value["{0}-vs-{1}".format(row['name1'], row['name2'])] \
            = row[v1]
        combo2value2["{0}-vs-{1}".format(row['name1'], row['name2'])] \
            = row[v2]

    table = defaultdict(list)
    samples = set(db['name1'].tolist()).union(set(db['name2'].tolist()))
    for samp1 in samples:
        for samp2 in samples:
            if samp1 == samp2:
                table['name1'].append(samp1)
                table['name2'].append(samp2)
                table[n1].append(1)
                table[n2].append(1)

            else:
                table['name1'].append(samp1)
                table['name2'].append(samp2)
                table[n1].append(np.nanmean([combo2value["{0}-vs-{1}".format(samp1,samp2)],
                            combo2value["{0}-vs-{1}".format(samp2,samp1)]]))
                table[n2].append(np.nanmean([combo2value2["{0}-vs-{1}".format(samp1,samp2)],
                            combo2value2["{0}-vs-{1}".format(samp2,samp1)]]))

    return pd.DataFrame(table)

def cluster_genome_strains(Mdb, kwargs):
    """
    Perform dRep-style clustering on the genomes
    """
    cluster_method = kwargs.get('clusterAlg')
    thresh = 1 - kwargs.get('ani_threshold')
    cov_thresh = kwargs.get('coverage_treshold')

    cdbs = []

    cluster_num = 1
    for genome, gdb in Mdb.groupby('genome'):
        # Average popANI values
        gdb = add_av_RC(gdb, v2='percent_compared', n2='av_cov')
        gdb['dist'] = 1 - gdb['av_ani']

        # Remove values with low coverage
        gdb['dist'] = [1 if c < cov_thresh else d for d, c in zip(gdb['dist'], gdb['av_cov'])]

        # Make squareform
        db = gdb.pivot("name1", "name2", 'dist')
        names = db.columns
        arr = np.asarray(db)
        arr = scipy.spatial.distance.squareform(arr, checks=True)

        # Cluster
        linkage = scipy.cluster.hierarchy.linkage(arr, method=cluster_method)
        fclust = scipy.cluster.hierarchy.fcluster(linkage, thresh,
                                                  criterion='distance')

        # Get Cdb
        cdb = _gen_cdb_from_fclust(fclust, names)
        cdb = cdb.rename(columns={'genome':'sample'})
        cdb['genome'] = genome

        # Parse cdb
        cdb['cluster'] = [f'{cluster_num}_{x}' for x in cdb['cluster']]
        cluster_num += 1

        # store
        cdbs.append(cdb)

    return pd.concat(cdbs).reset_index(drop=True)

def _gen_cdb_from_fclust(fclust,names):
    '''
    Make Cdb from the result of scipy.cluster.hierarchy.fcluster
    Args:
        fclust: result of scipy.cluster.hierarchy.fcluster
        names: list(db.columns) of the input dataframe
    Returns:
        DataFrame: Cdb
    '''
    Table={'cluster':[],'genome':[]}
    for i, c in enumerate(fclust):
        Table['cluster'].append(c)
        Table['genome'].append(names[i])

    return pd.DataFrame(Table)




import logging
import traceback
import pandas as pd
import seaborn as sns

from collections import defaultdict
import numpy as np

import matplotlib

matplotlib.use('Agg')
matplotlib.rcParams['pdf.fonttype'] = 42

import matplotlib.ticker as ticker
from matplotlib import pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.backends.backend_pdf import PdfPages

import inStrain.plotting.utilities
from inStrain.plotting.utilities import plot_genome
from inStrain.plotting.utilities import estimate_breadth

def linkage_decay_from_IS(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        db = IS.get('raw_linkage_table')
        db = db.sort_values('mm').drop_duplicates(subset=['scaffold', 'position_A', 'position_B'], keep='last')\
                    .sort_index().drop(columns=['mm'])

        stb = IS.get('scaffold2bin')
        Mdb = inStrain.genomeUtilities._add_stb(db, stb, verbose=False)
        assert len(Mdb) > 0
    except:
        logging.error("Skipping plot 5 - you don't have all required information. You need to run inStrain genome_wide first")
        traceback.print_exc()
        return

    # Make the plot
    logging.info("Plotting plot 5")
    name = 'LinkageDecay_plot.pdf'
    pp = PdfPages(plot_dir + name)

    for genome, mdb in Mdb.groupby('genome'):
        if not plot_genome(genome, IS, **kwargs):
            continue
        linkage_decay_plot(mdb, title=genome)
        fig = plt.gcf()
        fig.set_size_inches(6, 4)
        fig.tight_layout()
        pp.savefig(fig)#, bbox_inches='tight')
        #plt.show()
        plt.close(fig)

    # Save the figure
    pp.close()
    #plt.show()
    plt.close('all')

def linkage_decay_type_from_IS(IS, plot_dir=False, **kwargs):
    # Load the required data
    try:
        # Prepare
        db = IS.get('raw_linkage_table')
        db = db.sort_values('mm').drop_duplicates(subset=['scaffold', 'position_A', 'position_B'], keep='last')\
                    .sort_index().drop(columns=['mm'])
        stb = IS.get('scaffold2bin')
        Mdb = inStrain.genomeUtilities._add_stb(db, stb, verbose=False)

        SNdb = IS.get('SNP_mutation_types')
        assert SNdb is not None
        if len(SNdb) == 0:
            return
        SNdb.loc[:,'key'] = ["{0}:{1}".format(s, p) for s, p in zip(SNdb['scaffold'], SNdb['position'])]
        k2t = SNdb.set_index('key')['mutation_type'].to_dict()

        Mdb.loc[:,'link_type'] = Mdb.apply(calc_link_type, axis=1, k2t=k2t)
        assert len(Mdb) > 0
    except:
        logging.error("Skipping plot 8 - you don't have all required information. You need to run inStrain profile_genes first")
        if kwargs.get('debug', False):
            traceback.print_exc()
        return

    # Make the plot
    logging.info("Plotting plot 8")
    name = 'LinkageDecay_types_plot.pdf'
    pp = PdfPages(plot_dir + name)

    for genome, mdb in Mdb.groupby('genome'):
        if not plot_genome(genome, IS, **kwargs):
            continue
        db = linkage_decay_type(mdb, title=genome)
        fig = plt.gcf()
        fig.set_size_inches(6, 4)
        fig.tight_layout()
        pp.savefig(fig)#, bbox_inches='tight')
        #plt.show()
        plt.close(fig)

    # Save the figure
    pp.close()
    #plt.show()
    plt.close('all')

def linkage_decay_plot(db, chunkSize=5, min_vals=5, title=''):
    COLS = ['r2', 'r2_normalized', 'd_prime', 'd_prime_normalized']
    sns.set_style('white')

    # Make chunks
    max_d = db['distance'].max()
    table = defaultdict(list)
    numberChunks = max_d // chunkSize + 1
    #db['distance'] = db['distance'].astype(int)
    db = db.astype({'distance': int})
    for i in range(numberChunks):
        d = db[(db['distance'] >= int(i*chunkSize)) & (db['distance'] < int((i+1)*chunkSize))]

        table['d_start'].append(int(i*chunkSize))
        table['d_end'].append(int((i+1)*chunkSize))
        table['values'].append(len(d))
        for col in COLS:
            table[col].append(d[col].mean())
            table[col + '_values'].append(len(d[~d[col].isna()]))
    df = pd.DataFrame(table)
    df.loc[:,'distance'] = [np.mean([x, y]) for x, y in zip(df['d_start'], df['d_end'])]
    df = df.sort_values('distance')

    for col in COLS:
        df[col] = [c if v >= min_vals else np.nan for c, v in zip(df[col], df[col + '_values'])]

    for col in COLS:
        sns.lineplot(x=df['distance'], y=df[col], label=col, marker='o')

    plt.title(title)
    plt.xlabel('Distance between SNPs (bp)\nAveraged over {0}bp windows; plotting windows with at least {1} values'.format(
                chunkSize, min_vals))
    plt.ylabel("SNP linkage")
    return df

def calc_link_type(row, k2t):
    k1 = "{0}:{1}".format(row['scaffold'], row['position_A'])
    k2 = "{0}:{1}".format(row['scaffold'], row['position_B'])

    if (k1 in k2t) & (k2 in k2t):
        k1 = k2t[k1]
        k2 = k2t[k2]

        return "{0}-{1}".format(k1, k2)

    else:
        return np.nan

def linkage_decay_type(Odb, chunkSize=5, min_vals=2, title=''):
    sns.set_style('white')

    COLS = ['r2', 'r2_normalized', 'd_prime', 'd_prime_normalized']

    # Make chunks
    max_d = Odb['distance'].max()
    table = defaultdict(list)
    numberChunks = max_d // chunkSize + 1
    for lt in ['S=S', 'N-N', 'all']:
        if lt == 'all':
            db = Odb
        else:
            db = Odb[Odb['link_type'] == lt]

        #db['distance'] = db['distance'].astype(int)
        db = db.astype({'distance':int})
        for i in range(numberChunks):
            d = db[(db['distance'] >= int(i*chunkSize)) & (db['distance'] < int((i+1)*chunkSize))]

            table['d_start'].append(int(i*chunkSize))
            table['d_end'].append(int((i+1)*chunkSize))
            table['values'].append(len(d))
            table['link_type'].append(lt)
            for col in COLS:
                table[col].append(d[col].mean())
                table[col + '_values'].append(len(d[~d[col].isna()]))
    df = pd.DataFrame(table)
    df.loc[:,'distance'] = [np.mean([x, y]) for x, y in zip(df['d_start'], df['d_end'])]
    df = df.sort_values('distance')

    for thing in ['S-S', 'N-N', 'all']:
        if thing == 'all':
            fdb = df
        else:
            fdb = df[df['link_type'] == thing]
        sns.lineplot(data=fdb, x='distance', y='r2', label=thing, marker='o')

    plt.title(title)
    plt.xlabel('Distance between SNPs (bp)\nAveraged over {0}bp windows; plotting windows with at least {1} values'.format(
                chunkSize, min_vals))
    plt.ylabel("SNP linkage")
    return df


"""
Methods that make external system calls to samtools
"""

import os
import sys
import pysam
import logging
from subprocess import call
from Bio import bgzf


def prepare_bam_fie(bam, processes=6):
    """
    Make this a .bam file

    Args:
        bam = location of .bam file
    """
    if bam[-4:] == '.sam':
        logging.info("You gave me a sam- I'm going to make it a .bam now")

        bam = _sam_to_bam(bam)
        bam = _sort_index_bam(bam, processes=processes)

    elif bam[-4:] == '.bam':
        # If there's an index, assume its sorted
        if (os.path.exists(bam + '.bai')) | (os.path.exists(bam[:-4] + '.bai')):
            pass

        # If not sorted, sort and index
        elif not _is_sorted_bam(bam):
            bam = _sort_index_bam(bam, processes=processes, rm_ori=False)

        # If sorted, just index
        else:
            logging.info("Indexing {0}".format(bam))
            cmd = ['samtools', 'index', bam, bam + '.bai', '-@', str(processes)]
            print(' '.join(cmd))
            call(cmd)

    if os.stat(bam).st_size == 0:
        logging.error("Failed to generated a sorted .bam file! Make sure you have " +
                      "samtools version 1.6 or greater.")
        sys.exit()

    # Do a quick sanity check
    try:
        pysam.AlignmentFile(bam).mapped
    except ValueError:
        logging.error("It seems like the .bam file could not be indexed!" +
                      "Make sure you have samtools version 1.6 or greater.")
        sys.exit()

    return bam


def _sam_to_bam(sam):
    """
    From the location of a .sam file, convert it to a bam file and return the location
    """
    if sam[-4:] != '.sam':
        print('Sam file needs to end in .sam')
        sys.exit()

    bam = sam[:-4] + '.bam'
    logging.info("Converting {0} to {1}".format(sam, bam))
    cmd = ['samtools', 'view', '-S', '-b', sam, '>', bam]
    print(' '.join(cmd))
    call(' '.join(cmd), shell=True)

    return bam


def _sort_index_bam(bam, processes=6, rm_ori=False):
    """
    From a .bam file, sort and index it. Remove original if rm_ori
    Return path of sorted and indexed bam
    """
    if bam[-4:] != '.bam':
        logging.error('Bam file needs to end in .bam')
        sys.exit()

    if 'sorted.bam' not in bam:
        logging.info("sorting {0}".format(bam))
        sorted_bam = bam[:-4] + '.sorted.bam'
        cmd = ['samtools', 'sort', bam, '-o', sorted_bam, '-@', str(processes)]
        print(' '.join(cmd))
        call(cmd)
    else:
        sorted_bam = bam
        rm_ori = False

    logging.info("Indexing {0}".format(sorted_bam))
    cmd = ['samtools', 'index', sorted_bam, sorted_bam + '.bai', '-@', str(processes)]
    print(' '.join(cmd))
    call(cmd)

    if rm_ori:
        logging.info("Deleting {0}".format(bam))
        os.remove(bam)

    return sorted_bam


def _is_sorted_bam(bam):
    """
    Checks if a BAM file is sorted by coordinate.
    """
    with bgzf.BgzfReader(bam, "rb") as fin:
        bam_header = fin.readline().strip()
        return b"SO:coordinate" in bam_header



'''
This has utilities related to SNVs and SNSs
'''

import numpy as np
import pandas as pd

import inStrain
import inStrain.profile.profile_utilities

P2C = {'A':0, 'C':1, 'T':2, 'G':3} # base -> position
C2P = {0:'A', 1:'C', 2:'T', 3:'G'} # position -> base

def generate_snp_model(model_file, fdr=1e-6):
    '''
    The model_file contains the columns "coverage", "probability of X coverage base by random chance"

    The model that it returns has the properties:
        model[position_coverage] = the minimum number of alternate bases to be legit
        model[-1] = use this value if coverage is not in dictionary
    '''
    f = open(model_file)
    model = {}
    for line in f.readlines():
        if 'coverage' in line:
            continue
        counts = line.split()[1:]
        coverage = line.split()[0]
        i = 0 # note 6.4.20 - this should be 1, not zero
        for count in counts:
            if float(count) < fdr:
                model[int(coverage)] = i
                break
            i += 1

    model[-1] = max(model.values())

    return model

def update_snp_table(Stable, clonT, clonTR, MMcounts, p2c,\
        pos, scaff, mLen, ref_base, null_model, min_cov=5, min_covR=50, min_freq=.05):
    '''
    Add information to SNP table. Update basesCounted and snpsCounted

    Also update clonality

    cryptic means that it's a SNP at lower mm levels, and then not again at higher levels

    6.10.18 - v0.4.1 - Notes

    What makes it into the Stable?
    *  There must be two bases that satisfy the following conditions:
        *  There are more than min_cov number of bases counted
        *  There are more than min_freq percent of reads at the variant base
        *  The number of variant bases is more than the null model for that coverage
    *  If those are satisfied twice, the base with the highest count is stored as the "var_base",
       and the the bases with the second highest is the "ref_base"

    What is anySNP?
    *  For at least one mm position, there was a SNP called

    What is bases?
    *  A set of bases that are the var_base or ref_base at least once

    What are the things that are being missed?
    *  The case where theres only one base that meets the chriteria above, but its
       different from the reference base
    *  The case where >2 variant bases are present

    What is SNP table used for?
    *  Looking for specific variant positions; performing dn/ds analysis
    *  Establishing linkage patterns

    What to do about SNPs that are fixed opposite of reference?
    *  Create a "allele_count" column (formerly morphia); 2 means there are 2 bases; 1 means theres 1
        (and its opposite of reference); 3/4 means there multiple, though only
        two will be logged
        *  Right now only 2+ are being shown; so it will be easy to filter others out
    *  Throw in a check for if there's only one bases that fits that chriteria,
       is it not the reference base?
    *  Add "ref_base" back into the table to make it easier to compare SNP tables

    '''
    anySNP = False
    bases = set()
    ret_counts = None
    for mm in sorted(list(MMcounts.keys())):
        counts = inStrain.profile.profile_utilities.mm_counts_to_counts(MMcounts, mm)
        ret_counts = counts
        snp, morphia = call_snv_site(counts, ref_base, null_model, min_cov=min_cov, min_freq=min_freq) # Call SNP

        # Update clonality
        if mm not in clonT:
            clonT[mm] = np.full(mLen, fill_value=np.nan, dtype="float32")
        if sum(counts) >= min_cov:
            clonT[mm][pos] = calculate_clonality(counts)

        # Update rarefied clonality
        if mm not in clonTR:
            clonTR[mm] = np.full(mLen, fill_value=np.nan, dtype="float32")
        if sum(counts) >= min_covR:
            clonTR[mm][pos] = calculate_rarefied_clonality(counts, rarefied_coverage=min_covR)

        if not snp: # means base was not counted
            continue

        elif snp != -1: # means this is a SNP

           # calculate var_base
           counts_temp = list(counts)
           counts_temp[P2C[snp]] = 0
           var_base = C2P[list(counts_temp).index(sorted(counts_temp)[-1])] # this fixes the var_base = ref_base error when there's a tie - alexcc 5/8/2019

           SNP_class = calc_snp_class(snp, ref_base, var_base, counts,
                                        morphia, null_model, min_cov=min_cov,
                                        min_freq=min_freq)

           Stable['scaffold'].append(scaff)
           Stable['position'].append(pos)
           Stable['ref_base'].append(ref_base)
           for b, c in zip(['A', 'C', 'T', 'G'], counts):
               Stable[b].append(c)
           Stable['con_base'].append(snp)
           Stable['var_base'].append(var_base)
           Stable['mm'].append(mm)
           Stable['allele_count'].append(morphia)
           Stable['class'].append(SNP_class)

           if morphia >= 2:
               anySNP = True
               bases.add(snp)
               bases.add(var_base)
               ret_counts = counts

           elif (morphia == 1) & (anySNP == True):
               p2c[pos] = True

        # if it's now not a SNP, but it was in the past, mark it cryptic
        elif (snp == -1) & (anySNP == True):
            p2c[pos] = True

    if ret_counts is None:
        ret_counts = np.zeros(4, dtype=int)

    return anySNP, bases, ret_counts

def call_snv_site(counts, ref_base, model_to_use, min_cov=5, min_freq=0.05):
    '''
    Determines whether a site has a variant based on its nucleotide count frequencies.

    Returns one of the following and the "morphia" (number of bases present in the reads):
        Base if SNP
        -1 if not SNP
        None if unCounted

    A base is considered present in the reads if:
    *  There are more than min_cov number of bases counted
    *  There are more than min_freq percent of reads at the variant base
    *  The number of variant bases is more than the null model for that coverage

    The morphia is the number of bases present in the reads. Things are returned as SNPs if:
    *  Morphia >= 2
    *  Morphia == 1, and the base present is not the genome reference base
    *  Morphia == 0 (this is a special case meaning no bases are really present)
    '''
    # Make sure you have the coverage
    total = sum(counts)
    if total < min_cov:
        return None, 0

    # Count how many bases are there
    i = 0
    for c in counts:
        if total in model_to_use:
            min_bases = model_to_use[total]
        else:
            min_bases = model_to_use[-1]

        if c >= min_bases and float(c) / total >= min_freq:
            i += 1

    # If there are 2, say that
    if i > 1:
        return C2P[np.argmax(counts)], i

    # If theres only 1, see if its the reference base
    elif (i == 1) & (C2P[np.argmax(counts)] != ref_base):
        return C2P[np.argmax(counts)], i

    # That means this is a super polymorphic position with no dominant bases
    elif i == 0:
        return C2P[np.argmax(counts)], i

    # This means its just not a SNP; one dominant reference base
    else:
        return -1, i

def calc_snp_class(con_base, ref_base, var_base, counts, allele_count, null_model, min_cov=5, min_freq=0.05):
    '''
    Return the class of SNP this is
    '''
    if ref_base not in ['A', 'C', 'T', 'G']:
        return 'AmbiguousReference'

    if allele_count == 0:
        return 'DivergentSite'

    elif allele_count == 1:
        assert con_base != ref_base
        return 'SNS'

    else:
        if ref_base == con_base:
            return 'SNV'

        else:
            if ref_base == var_base:
                return 'con_SNV'

            if inStrain.readComparer.is_present(counts[P2C[ref_base]], sum(counts), null_model, float(min_freq)):
                return 'con_SNV'

            return 'pop_SNV'

def calculate_clonality(counts):
    '''
    Calculates the probability that two reads have the same allele at a position (per site nucleotide diversity)
    '''
    s = sum(counts)
    prob = (float(counts[0]) / s) * (float(counts[0]) / s) + (float(counts[1]) / s) * (float(counts[1]) / s) + (float(counts[2]) / s) * (float(counts[2]) / s) + (float(counts[3]) / s) * (float(counts[3]) / s)
    return prob

def calculate_rarefied_clonality(counts, rarefied_coverage=50):
    '''
    Calculates the probability that two reads have the same allele at a position (per site nucleotide diversity)
    '''
    # Figure out the probability distribution
    s = sum(counts)
    p = [counts[i]/s for i in [0,1,2,3]]

    # Make rarefied counts
    rcounts_list = np.random.choice([0,1,2,3], rarefied_coverage, p=p)
    unique, ncounts = np.unique(rcounts_list, return_counts=True)
    item2counts = dict(zip(unique, ncounts))
    rcounts = [item2counts[i] if i in item2counts else 0 for i in [0,1,2,3]]

    return calculate_clonality(rcounts)

def calc_snps(Odb, mm, null_model, min_freq=0.05):
    '''
    Calculate the number of reference SNPs, bi-allelic SNPs, multi-allelic SNPs (>2), total SNPs, consensus_SNPs, and poplation_SNPs

    Vocabulary:

    SNS_count = (number of substitutions) (allele_count = 1 and ref != con)
    SNV_count = number of SNVs (allele_count > 1)

    con_snps = con != ref and allele count > 0
    pop_snps = confusing
    '''
    if len(Odb) == 0:
        return [0, 0, 0, 0, 0]

    db = Odb[Odb['mm'] <= mm].sort_values('mm').drop_duplicates(subset=['position'], keep='last')

    SNS_count = len(db[(db['allele_count'] == 1)])
    SNV_count = len(db[db['allele_count'] > 1])

    con_snps = len(db[db['class'].isin(['SNS', 'con_SNV', 'pop_SNV'])])
    pop_snps = len(db[db['class'].isin(['SNS', 'pop_SNV'])])

    return [SNS_count, SNV_count, len(db), con_snps, pop_snps]

def generate_snp_table(Stable, scaffold, p2c):
    '''
    Make a dataframe and parse it a little bit
    '''
    SNPTable = pd.DataFrame(Stable)
    if len(SNPTable) > 0:
        SNPTable['scaffold'] = scaffold

        # Add cryptic SNPs
        SNPTable.loc[:,'cryptic'] = SNPTable['position'].map(p2c)
        SNPTable.loc[:,'cryptic'] = SNPTable['cryptic'].fillna(False)

        # Calc base coverage
        SNPTable.loc[:,'position_coverage'] = [sum([a,c,t,g]) for a,c,t,g in zip(SNPTable['A'],SNPTable['C'],SNPTable['T'],SNPTable['G'])]

    del Stable
    return SNPTable



import sys
import logging
import numpy as np
from matplotlib import pyplot as plt

import inStrain
import inStrain.genomeUtilities

def plot_genome(genome, IS, **kwargs):
    '''
    Decide if this genome should be plotted based on the filters in the kwargs. Return True if so

    GWdb is the result of Mdb = inStrain.genomeUtilities.genomeWideFromIS(IS, 'scaffold_info', mm_level=True)
    '''
    GWdb = kwargs.get('GWdb', False)

    # FILTER BY GENOME LIST
    genomes = kwargs.get('genomes', None)
    if genomes is not None:
        if genome not in genomes:
            return False
        else:
            return True

    # FILTER BY BREADTH
    mb = float(kwargs.get('minimum_breadth', 0))
    if mb > 0:
        if GWdb is False:
            GWdb = inStrain.genomeUtilities.genomeWideFromIS(IS, 'scaffold_info', mm_level=False)
        if 'mm' in GWdb.columns:
            GWdb = GWdb.sort_values('mm').drop_duplicates(subset='genome', keep='last')
        try:
            breadth = GWdb[GWdb['genome'] == genome]['breadth'].tolist()[0]
        except:
            breadth = 0
        if float(breadth) < mb:
            return False

    return True

def estimate_breadth(coverage):
    '''
    Estimate breadth based on coverage

    Based on the function breadth = -1.000 * e^(0.883 * coverage) + 1.000
    '''
    return (-1) * np.exp(-1 * ((0.883) * coverage)) + 1

def _calc_ylim(max_size):
    return min(max(5, max_size/5000), 100)

def _shorten_name(name):
    name = name.replace('.sorted.bam', '')
    if '-vs-' in name:
        name = name.split('-vs-')[1]
    if len(name) > 15:
        name = '\n'.join(name[n:n + 15] for n in range(0, len(name), 15))
    return name

def gen_color_dictionary(names, name2cluster):
    '''
    Make the dictionary name2color
    Args:
        names: key in the returned dictionary
        name2cluster: a dictionary of name to it's cluster
    Returns:
        dict: name -> color
    '''
    #cm = _rand_cmap(len(set(name2cluster.values()))+1,type='bright')
    vals = np.linspace(0,1,len(set(name2cluster.values()))+1)
    np.random.shuffle(vals)
    cm = plt.cm.colors.ListedColormap(plt.cm.jet(vals))

    # 1. generate cluster to color
    cluster2color = {}
    clusters = set(name2cluster.values())
    NUM_COLORS = len(clusters)
    for cluster in clusters:
        try:
            cluster2color[cluster] = cm(1.*int(cluster)/NUM_COLORS)
        except:
            cluster2color[cluster] = cm(1.*int(str(cluster).split('_')[1])/NUM_COLORS)

    #2. name to color
    name2color = {}
    for name in names:
        name2color[name] = cluster2color[name2cluster[name]]

    return name2color

def _x_fig_size(points, factor= .07, min= 8):
    '''
    Calculate how big the x of the figure should be
    '''
    size = points * factor
    return max([size,min])


import os
import sys
import logging
import traceback

import inStrain
import inStrain.SNVprofile
import inStrain.plotting
import inStrain.plotting.utilities

from inStrain.plotting.mapping_plots import mm_plot_from_IS, ANI_dist_plot_from_IS, read_filtering_from_IS
from inStrain.plotting.positional_plots import genome_plot_from_IS, scaffold_inspection_from_IS
from inStrain.plotting.SNV_plots import allele_freq_plot_from_IS
from inStrain.plotting.gene_plots import gene_histogram_from_IS
from inStrain.plotting.linkage_plots import linkage_decay_from_IS, linkage_decay_type_from_IS
from inStrain.plotting.compare_plots import dendrograms_from_RC

class PlottingController(object):
    """
    Handle the logic of profiling
    """
    PLOTTING_FUNCTIONS = [
        [1, mm_plot_from_IS],
        [2, genome_plot_from_IS],
        [3, ANI_dist_plot_from_IS],
        [4, allele_freq_plot_from_IS],
        [5, linkage_decay_from_IS],
        [6, read_filtering_from_IS],
        [7, scaffold_inspection_from_IS],
        [8, linkage_decay_type_from_IS],
        [9, gene_histogram_from_IS],
        [10, dendrograms_from_RC]
    ]

    def __init__(self, args):
        """
        Accepts args directly from argparse
        """
        self.ori_args = args
        self.args = args

        self.validate_input()

    def main(self):
        """
        Profile the bam with inStrain using the options in kwargs
        """
        if not self.validated:
            return

        # Establish kwargs, which is kind of like the cacheing of this module
        self.establish_kwargs()

        # Run the plots
        self.run_plots()

    def run_plots(self):
        """
        Call to create all the plots
        """
        for plf in PlottingController.PLOTTING_FUNCTIONS:
            number, func = plf
            if str(number) in self.to_plot:
                try:
                    func(self.IS, plot_dir=self.plot_dir, **self.kwargs)
                except BaseException as e:
                    logging.error(f'Failed to make plot #{number}: {str(e)}')
                    if self.debug:
                        traceback.print_exc()
                        logging.debug(traceback.format_exc())


    def establish_kwargs(self):
        """
        Load data from the IS file that will be based to plots through the kwargs
        """
        # Remove IS from the kwargs
        kwargs = vars(self.args)
        kwargs.pop('IS')

        # Cache needed data
        if self.IS_TYPE == 'IS':
            try:
                kwargs['GWdb'] = self.IS.get('genome_level_info')
            except:
                logging.error(
                    "Cannot cache scaffold info - you don't have all required information. You need to run inStrain genome_wide first")
                if self.debug:
                    traceback.print_exc()

        if (('2' in self.to_plot) | ('7' in self.to_plot)):
            kwargs['covT'] = self.IS.get('covT')
            kwargs['clonT'] = self.IS.get('clonT')
            kwargs['raw_linkage_table'] = self.IS.get('raw_linkage_table')
            kwargs['cumulative_snv_table'] = self.IS.get('cumulative_snv_table')

        self.kwargs = kwargs

    def validate_input(self):
        """
        Do a bunch of parsing of the input arguments
        """
        args = self.args
        kwargs = vars(args)

        # Get the IS object
        assert os.path.exists(args.IS)
        self.IS = inStrain.SNVprofile.SNVprofile(args.IS)

        # Set up the logger
        log_loc = self.IS.get_location('log') + 'log.log'
        inStrain.controller.setup_logger(log_loc)
        logging.getLogger('matplotlib.font_manager').disabled = True

        # Set up the .stb
        stb = self.IS.get('scaffold2bin')
        if stb is None:
            logging.error("This IS object does not have an .stb file; cant use it to make plots")
            self.validated = False

        # Set debug
        self.debug = args.debug

        # Figure out if this is an RC or a IS
        bl = self.IS.get('bam_loc')
        if bl is None:
            self.IS_TYPE = 'RC'
            self.options = ['10']
        else:
            self.IS_TYPE = 'IS'
            self.options = ['1', '2', '3', '4', '5', '6', '7', '8', '9']

        # Get the plot directory and basename
        self.plot_dir = self.IS.get_location('figures') + os.path.basename(self.IS.get('location')) + '_'

        # Figure out which plots to make
        self.to_plot = _parse_plot_options(self.options, kwargs.get('plots', None))
        logging.info("making plots {0}".format(', '.join(self.to_plot)))
        self.validated = True

def _parse_plot_options(options, args):
    '''
    Read user input and figure out a list of plots to make

    Args:
        options: list of possible plots to make (default [1-6])
        args: the command line passed in

    Returns:
        list: list of ints in the args
    '''
    to_plot = []

    if args[0] in ['all', 'a']:
        to_plot += options

    elif args == None:
        logging.error("No plots given!")
        return []

    else:
        for arg in args:
            if arg in options:
                to_plot.append(arg)

    return to_plot




'''
This file contains the API-like elements that are called from the outside
'''

# import inStrain.profile.profile_controller
#
# def profile_bam(bam, fasta_db, sR2M, ISP_loc, **kwargs):
#     '''
#     Profile a bam file with inStrain
#
#     arguments:
#         bam = location of .bam file
#         Fdb = dictionary listing fasta locations to profile
#         sR2M = dictionary of scaffold -> read pair -> number of mm
#         ISP_loc = location of inStrain profile to store results
#     '''
#     return inStrain.profile.profile_controller.BamProfileController(
#                                         bam, fasta_db, sR2M, ISP_loc, **kwargs).main()



"""
I tried to go down a seaborn-inspired structure here, but gave up on it. Leaving here in case want to go down
this road again in the future.

- MO 2/3/21
"""


import logging
import traceback

import matplotlib
matplotlib.use('Agg')
import matplotlib.ticker as ticker
from matplotlib import pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.backends.backend_pdf import PdfPages

import inStrain.plotting
import inStrain.plotting.utilities

def BreadthCurvePlot(IS, **kwargs):
    """
    Create a breadth curve plot (plot number 1)
    """
    plotter = _BreadthCurvePlotter(IS, **kwargs)

    logging.info("Plotting plot 1")
    plotter.plot()

class _BreadthCurvePlotter:
    """
    Draw the breadth curve plot
    """
    def plot(self):
        logging.info("Plotting plot 1")

        # Establish multipage .pdf
        pp = PdfPages(self.plot_dir + self.name)

        # Make the plots
        for genome, mdb in self.GWdb.groupby('genome'):

            # See if you should plot this genome
            if not inStrain.plotting.utilities.plot_genome(genome, self.IS, **self.kwargs):
                continue

    def __init__(self, IS, **kwargs):
        """
        Initialize the plotting object
        """
        # Load simple kwargs
        self.debug = kwargs.get('debug', False)
        self.plot_dir = kwargs.get('plot_dir', None)
        self.name = kwargs.get('name', 'CoverageAndBreadth_vs_readMismatch.pdf')
        self.kwrags = kwargs
        self.IS = IS

        # Load data
        try:
            # Load GWdb cache
            if kwargs.get('GWdb', None) is not None:
                GWdb = kwargs.get('GWdb')
            else:
                GWdb = None
                print('Make GWdb you fool')
            assert len(GWdb) > 0

            # Parse GWdb a bit
            readLen = int(IS.get_read_length())
            GWdb['read_length'] = readLen
            GWdb['mm'] = GWdb['mm'].astype(int)
            GWdb['ANI_level'] = [(readLen - mm) / readLen for mm in GWdb['mm']]

        except:
            logging.error(
                "Skipping plot 1 - you don't have all required information.")
            traceback.print_exc()
            return




#!/usr/bin/env python

import os
import csv
import sys
import time
import glob
import logging
import warnings
import argparse
import traceback
import multiprocessing

import Bio
import numpy as np
import pandas as pd
from tqdm import tqdm
from Bio import SeqIO
import concurrent.futures
from concurrent import futures
from inStrain import SNVprofile
from collections import defaultdict

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from Bio.codonalign.codonalphabet import default_codon_table

import inStrain.SNVprofile
import inStrain.controller
#import inStrain.profileUtilities
import inStrain.logUtils

class Controller():
    '''
    The command line access point to the program
    '''
    def main(self, args):
        '''
        The main method when run on the command line
        '''
        # Parse arguments
        args = self.validate_input(args)

        vargs = vars(args)
        IS = vargs.pop('IS')
        GF = vargs.pop('gene_file')

        # Read the genes file
        logging.debug('Loading genes')
        scaff_2_gene_database, scaff2gene2sequence = parse_genes(GF, **vargs)
        GdbP = pd.concat([x for x in scaff_2_gene_database.values()])

        # Calculate all your parallelized gene-level stuff
        name2result = calculate_gene_metrics(IS, GdbP, scaff2gene2sequence, **vargs)

        # Store information
        IS.store('genes_fileloc', GF, 'value', 'Location of genes file that was used to call genes')
        IS.store('genes_table', GdbP, 'pandas', 'Location of genes in the associated genes_file')
        IS.store('genes_coverage', name2result['coverage'], 'pandas', 'Coverage of individual genes')
        IS.store('genes_clonality', name2result['clonality'], 'pandas', 'Clonality of individual genes')
        IS.store('genes_SNP_count', name2result['SNP_density'], 'pandas', 'SNP density and counts of individual genes')
        IS.store('SNP_mutation_types', name2result['SNP_mutation_types'], 'pandas', 'The mutation types of SNPs')

        if vargs.get('store_everything', False):
            IS.store('scaff2gene2sequence', scaff2gene2sequence, 'pickle', 'Dicitonary of scaff -> gene -> nucleotide sequence')

        # Store the output
        IS.generate('gene_info')
        IS.generate("SNVs")

    def validate_input(self, args):
        '''
        Validate and mess with the arguments a bit
        '''
        # Make sure the IS object is OK and load it
        assert os.path.exists(args.IS)
        args.IS = inStrain.SNVprofile.SNVprofile(args.IS)

        # Set up the logger
        log_loc = args.IS.get_location('log') + 'log.log'
        inStrain.controller.setup_logger(log_loc)

        return args

def gene_profile_worker(gene_cmd_queue, gene_result_queue, single_thread=False):
    '''
    Worker to profile splits
    '''
    while True:
        # Get command
        if not single_thread:
            cmds = gene_cmd_queue.get(True)
        else:
            try:
                cmds = gene_cmd_queue.get(timeout=5)
            except:
                return

        # Process cmd
        GPs = profile_genes_wrapper(cmds)
        gene_result_queue.put(GPs)

def profile_genes_wrapper(cmds):
    '''
    Take a group of commands and run geneprofile
    '''
    results = []
    for cmd in cmds:
        try:
            results.append(profile_genes(cmd.scaffold, **cmd.arguments))
        except Exception as e:
            print(e)
            traceback.print_exc()
            logging.error("FAILURE GeneException {0}".format(str(cmd.scaffold)))
            results.append(None)
    return results

def calculate_gene_metrics(IS, GdbP, scaff2gene2sequenceP, **kwargs):
    '''
    Calculate the metrics of all genes on a parallelized scaffold-level basis

    IS = Initialized inStrain.SNVprofile
    GdbP = List of gene locations
    gene2sequenceP = Dicitonary of gene -> nucleotide sequence
    '''
    inStrain.logUtils.log_checkpoint("GeneProfile", "calculate_gene_metrics", "start")

    # Get key word arguments for the wrapper
    p = int(kwargs.get('processes', 6))

    # Make a list of scaffolds to profile the genes of
    scaffolds_with_genes = set(GdbP['scaffold'].unique())
    scaffolds_in_IS = set(IS._get_covt_keys())
    scaffolds_to_profile = scaffolds_with_genes.intersection(scaffolds_in_IS)
    logging.info("{0} scaffolds with genes in the input; {1} scaffolds in the IS, {2} to compare".format(
            len(scaffolds_with_genes), len(scaffolds_in_IS), len(scaffolds_to_profile)))

    # Calculate scaffold -> number of genes to profile
    s2g = GdbP['scaffold'].value_counts().to_dict()
    kwargs['s2g'] = s2g

    # Make global objects for the profiling
    inStrain.logUtils.log_checkpoint("GeneProfile", "make_globals", "start")
    global CumulativeSNVtable
    CumulativeSNVtable = IS.get('cumulative_snv_table')
    if len(CumulativeSNVtable) > 0:
        CumulativeSNVtable = CumulativeSNVtable.sort_values('mm')
    else:
        CumulativeSNVtable = pd.DataFrame(columns=['scaffold'])

    global covTs
    covTs = IS.get('covT', scaffolds=scaffolds_to_profile)

    global clonTs
    clonTs = IS.get('clonT', scaffolds=scaffolds_to_profile)

    global scaff2gene2sequence
    scaff2gene2sequence =  scaff2gene2sequenceP

    global Gdb
    Gdb = GdbP
    inStrain.logUtils.log_checkpoint("GeneProfile", "make_globals", "end")

    # Generate commands and queue them
    logging.debug('Creating commands')
    cmd_groups = [x for x in iterate_commands(scaffolds_to_profile, Gdb, kwargs)]
    logging.debug('There are {0} cmd groups'.format(len(cmd_groups)))

    inStrain.logUtils.log_checkpoint("GeneProfile", "create_queue", "start")
    gene_cmd_queue = multiprocessing.Queue()
    gene_result_queue = multiprocessing.Queue()
    GeneProfiles = []

    for cmd_group in cmd_groups:
        gene_cmd_queue.put(cmd_group)

    inStrain.logUtils.log_checkpoint("GeneProfile", "create_queue", "end")

    if p > 1:
        logging.debug('Establishing processes')
        processes = []
        for i in range(0, p):
            processes.append(multiprocessing.Process(target=gene_profile_worker, args=(gene_cmd_queue, gene_result_queue)))
        for proc in processes:
            proc.start()

        # Set up progress bar
        pbar = tqdm(desc='Profiling genes: ', total=len(cmd_groups))

        # Get the results
        recieved_profiles = 0
        while recieved_profiles < len(cmd_groups):
            GPs = gene_result_queue.get()
            recieved_profiles += 1
            pbar.update(1)
            for GP in GPs:
                if GP is not None:
                    logging.debug(GP[4])
                    GeneProfiles.append(GP)

        # Close multi-processing
        for proc in processes:
            proc.terminate()

        # Close progress bar
        pbar.close()

    else:
        gene_profile_worker(gene_cmd_queue, gene_result_queue, single_thread=True)
        logging.info("Done profiling genes")

        # Get the genes
        recieved_profiles = 0
        while recieved_profiles < len(cmd_groups):
            logging.debug('going to grab at {0}'.format(recieved_profiles))
            GPs = gene_result_queue.get(timeout=5)
            logging.debug('did a grab at {0}'.format(recieved_profiles))
            recieved_profiles += 1
            for GP in GPs:
                if GP is not None:
                    logging.debug(GP[4])
                    GeneProfiles.append(GP)

    inStrain.logUtils.log_checkpoint("GeneProfile", "return_results", "start")
    name2result = {}
    for i, name in enumerate(['coverage', 'clonality', 'SNP_density', 'SNP_mutation_types']):
        name2result[name] = pd.concat([G[i] for G in GeneProfiles])
    inStrain.logUtils.log_checkpoint("GeneProfile", "return_results", "end")

    inStrain.logUtils.log_checkpoint("GeneProfile", "calculate_gene_metrics", "end")
    return name2result

def profile_genes(scaffold, **kwargs):
    '''
    This is the money that gets multiprocessed

    Relies on having a global "Gdb", "gene2sequence", "CumulativeSNVtable", "covTs", and "clonTs"

    * Calculate the clonality, coverage, linkage, and SNV_density for each gene
    * Determine whether each SNP is synynomous or nonsynonymous
    '''
    # Log
    pid = os.getpid()
    log_message = "\nSpecialPoint_genes {0} PID {1} whole start {2}".format(scaffold, pid, time.time())

    # For testing purposes
    if ((scaffold == 'FailureScaffoldHeaderTesting')):
        assert False

    # Get the list of genes for this scaffold
    gdb = Gdb[Gdb['scaffold'] == scaffold]

    # Calculate gene-level coverage
    log_message += "\nSpecialPoint_genes {0} PID {1} coverage start {2}".format(scaffold, pid, time.time())
    if scaffold not in covTs:
        logging.info("{0} isnt in covT!".format(scaffold))
        cdb = pd.DataFrame()
    else:
        covT = covTs[scaffold]
        cdb = calc_gene_coverage(gdb, covT)
        del covT
    log_message += "\nSpecialPoint_genes {0} PID {1} coverage end {2}".format(scaffold, pid, time.time())

    # Calculate gene-level clonality
    log_message += "\nSpecialPoint_genes {0} PID {1} clonality start {2}".format(scaffold, pid, time.time())
    if scaffold not in clonTs:
        logging.info("{0} isnt in clovT!".format(scaffold))
        cldb = pd.DataFrame()
    else:
        clonT = clonTs[scaffold]
        cldb = calc_gene_clonality(gdb, clonT)
        del clonT
    log_message += "\nSpecialPoint_genes {0} PID {1} clonality end {2}".format(scaffold, pid, time.time())

    # Determine whether SNPs are synonmous or non-synonmous
    log_message += "\nSpecialPoint_genes {0} PID {1} SNP_character start {2}".format(scaffold, pid, time.time())
    Ldb = CumulativeSNVtable[CumulativeSNVtable['scaffold'] == scaffold]
    if len(Ldb) == 0:
        sdb = pd.DataFrame()
    else:
        sdb = Characterize_SNPs_wrapper(Ldb, gdb, scaff2gene2sequence[scaffold])
    log_message += "\nSpecialPoint_genes {0} PID {1} SNP_character end {2}".format(scaffold, pid, time.time())

    # Calculate gene-level SNP counts
    log_message += "\nSpecialPoint_genes {0} PID {1} SNP_counts start {2}".format(scaffold, pid, time.time())
    if len(Ldb) == 0:
        ldb = pd.DataFrame()
        sublog = ''
    else:
        #ldb = calc_gene_snp_density(gdb, Ldb)
        ldb, sublog = calc_gene_snp_counts(gdb, Ldb, sdb, scaff2gene2sequence[scaffold], scaffold=scaffold)
    log_message += "\nSpecialPoint_genes {0} PID {1} SNP_counts end {2}".format(scaffold, pid, time.time())
    log_message += sublog

    log_message += "\nSpecialPoint_genes {0} PID {1} whole end {2}".format(scaffold, pid, time.time())

    results = (cdb, cldb, ldb, sdb, log_message)

    return results

def profile_genes_from_profile(scaffold, gdb, covT, clonT, Ldb, gene2sequence):
    '''
    Call profile genes from elsewhere

    Arguments:
        scaffold = name of scaffold
        gdb = gene_datatable
        covT = covT for this scaffold
        clonT = clonT for this scaffold
        Ldb = cumulative SNP table for this scaffold

    * Calculate the clonality, coverage, linkage, and SNV_density for each gene
    * Determine whether each SNP is synynomous or nonsynonymous
    '''
    # Log
    log = inStrain.logUtils.get_worker_log('ProfileGenes', scaffold, 'start')

    # For testing purposes
    if ((scaffold == 'FailureScaffoldHeaderTesting')):
        assert False

    # Calculate gene-level coverage
    #log_message += "\nSpecialPoint_genes {0} PID {1} coverage start {2}".format(scaffold, pid, time.time())
    cdb = calc_gene_coverage(gdb, covT)
    #log_message += "\nSpecialPoint_genes {0} PID {1} coverage end {2}".format(scaffold, pid, time.time())

    # Calculate gene-level clonality
    # log_message += "\nSpecialPoint_genes {0} PID {1} clonality start {2}".format(scaffold, pid, time.time())
    cldb = calc_gene_clonality(gdb, clonT)
    # log_message += "\nSpecialPoint_genes {0} PID {1} clonality end {2}".format(scaffold, pid, time.time())

    # Determine whether SNPs are synonmous or non-synonmous
    #log_message += "\nSpecialPoint_genes {0} PID {1} SNP_character start {2}".format(scaffold, pid, time.time())
    sdb = Characterize_SNPs_wrapper(Ldb, gdb, gene2sequence)
    #log_message += "\nSpecialPoint_genes {0} PID {1} SNP_character end {2}".format(scaffold, pid, time.time())

    # Calculate gene-level SNP counts
    #log_message += "\nSpecialPoint_genes {0} PID {1} SNP_counts start {2}".format(scaffold, pid, time.time())
    ldb, sublog = calc_gene_snp_counts(gdb, Ldb, sdb, gene2sequence, scaffold=scaffold)
    #log_message += "\nSpecialPoint_genes {0} PID {1} SNP_counts end {2}".format(scaffold, pid, time.time())
    #log_message += sublog

    log += inStrain.logUtils.get_worker_log('ProfileGenes', scaffold, 'end')

    results = (cdb, cldb, ldb, sdb, log)

    return results

def calc_gene_coverage(gdb, covT):
    '''
    Gene-level and mm-level coverage
    '''
    table = defaultdict(list)

    for mm, cov in iterate_covT_mms(covT):
        if len(cov) == 0:
            continue

        for i, row in gdb.iterrows():
            gcov = cov.loc[int(row['start']):int(row['end'])]
            gLen = abs(row['end'] - row['start']) + 1

            table['gene'].append(row['gene'])
            table['coverage'].append(gcov.sum() / gLen)
            table['breadth'].append(len(gcov) / gLen)
            table['mm'].append(mm)

    return pd.DataFrame(table)

def iterate_clonT_mms(clonT):
    p2c = {}
    mms = sorted([int(mm) for mm in list(clonT.keys())])
    for mm in mms:
        for pos, val in clonT[mm].items():
            p2c[pos] = val

        inds = []
        vals = []
        for ind in sorted(p2c.keys()):
            inds.append(ind)
            vals.append(p2c[ind])

        yield mm, pd.Series(data = vals, index = np.array(inds).astype('int'))

def iterate_covT_mms(clonT):
    counts = pd.Series()
    mms = sorted([int(mm) for mm in list(clonT.keys())])
    for mm in mms:
        count = clonT[mm]
        counts = counts.add(count, fill_value=0)
        yield mm, counts

def calc_gene_clonality(gdb, clonT):
    '''
    Gene-level and mm-level clonality
    '''
    table = defaultdict(list)

    for mm, cov in iterate_clonT_mms(clonT):
        if len(cov) == 0:
            continue

        for i, row in gdb.iterrows():
            gcov = cov.loc[int(row['start']):int(row['end'])]
            gLen = abs(row['end'] - row['start']) + 1

            table['gene'].append(row['gene'])

            try:
                microdiversity = 1 - gcov.mean()
            except :
                microdiversity = np.nan

            #table['clonality'].append(gcov.mean())
            table['nucl_diversity'].append(microdiversity)
            table['breadth_minCov'].append(len(gcov) / gLen)
            table['mm'].append(mm)

    return pd.DataFrame(table)
#
# def calc_gene_snp_density(gdb, ldb):
#     '''
#     Gene-level and mm-level clonality
#     '''
#     table = defaultdict(list)
#
#     for mm in sorted(ldb['mm'].unique()):
#         db = ldb[ldb['mm'] <= mm].drop_duplicates(subset=['scaffold', 'position'], keep='last')
#         cov = db.set_index('position')['ref_base'].sort_index()
#         if len(cov) == 0:
#             continue
#
#         for i, row in gdb.iterrows():
#             gcov = cov.loc[int(row['start']):int(row['end'])]
#             gLen = abs(row['end'] - row['start']) + 1
#
#             table['gene'].append(row['gene'])
#             table['SNPs_per_bp'].append(len(gcov) / gLen)
#             table['mm'].append(mm)
#
#     return pd.DataFrame(table)



def count_sites(seq, k=1, codon_table=None):
    '''
    From a nucleotide sequence and codon table, calculate S and N sites
    '''
    codon_lst = convert_to_codons(seq)

    if codon_table is None:
        codon_table = default_codon_table
    S_site = 0.0  # synonymous sites
    N_site = 0.0  # non-synonymous sites
    purine = ('A', 'G')
    pyrimidine = ('T', 'C')
    base_tuple = ('A', 'T', 'C', 'G')
    for codon in codon_lst:
        neighbor_codon = {'transition': [], 'transversion': []}
        # classify neighbor codons
        codon = codon.replace('U', 'T')
        if codon == '---':
            continue
        if 'N' in codon:
            continue
        for n, i in enumerate(codon):
            for j in base_tuple:
                if i == j:
                    pass
                elif i in purine and j in purine:
                    codon_chars = [c for c in codon]
                    codon_chars[n] = j
                    this_codon = ''.join(codon_chars)
                    neighbor_codon['transition'].append(this_codon)
                elif i in pyrimidine and j in pyrimidine:
                    codon_chars = [c for c in codon]
                    codon_chars[n] = j
                    this_codon = ''.join(codon_chars)
                    neighbor_codon['transition'].append(this_codon)
                else:
                    codon_chars = [c for c in codon]
                    codon_chars[n] = j
                    this_codon = ''.join(codon_chars)
                    neighbor_codon['transversion'].append(this_codon)
        # count synonymous and non-synonymous sites
        #codon = codon.replace('T', 'U')
        if (codon in ['TAG', 'TAA', 'TGA']):
            #print("STOP DETECTED")
            continue
        aa = codon_table.forward_table[codon]
        this_codon_N_site = this_codon_S_site = 0
        for neighbor in neighbor_codon['transition']:
            if neighbor in codon_table.stop_codons:
                this_codon_N_site += 1
            elif codon_table.forward_table[neighbor] == aa:
                this_codon_S_site += 1
            else:
                this_codon_N_site += 1
        for neighbor in neighbor_codon['transversion']:
            if neighbor in codon_table.stop_codons:
                this_codon_N_site += k
            elif codon_table.forward_table[neighbor] == aa:
                this_codon_S_site += k
            else:
                this_codon_N_site += k
        norm_const = (this_codon_N_site + this_codon_S_site)/3
        S_site += float(this_codon_S_site) / float(norm_const)
        N_site += float(this_codon_N_site) / float(norm_const)
    return (S_site, N_site)

def convert_to_codons(seq):
    codons = []
    for c in zip(*(iter(seq),) * 3):
        co = ''.join(c)
        assert len(co) == 3
        codons.append(co)
    return codons

def calc_gene_snp_counts(gdb, ldb, sdb, gene2sequence, scaffold=None):
    '''
    Count the number of SNPs in each gene, as well as N and S sites

    RELIES ON HAVING gene2sequence AS A GLOBAL (needed for multiprocessing speed)

    Argumnets:
        gdb = table of genes
        ldb = Raw cumulative snp table for a single scaffold (mm-level)
        sdb = SNP table with N and S and I annotated
    '''
    if len(ldb) == 0:
        return pd.DataFrame(), ''

    pid = os.getpid()

    # Merge ldb and sdb
    xdb = pd.merge(ldb, sdb[['position', 'mutation_type', 'gene']],
            on=['position'], how='left').reset_index(drop=True)

    # Calculate counts of N and S sites
    log_message = "\nSpecialPoint_genes {0} PID {1} SNP_counts_SiteCalc start {2}".format(scaffold, pid, time.time())
    table = defaultdict(list)
    for i, row in gdb.iterrows():
        try:
            S_site, N_site = count_sites(gene2sequence[row['gene']])
        except:
            S_site = np.nan
            N_site = np.nan

        table['gene'].append(row['gene'])
        table['S_sites'].append(S_site)
        table['N_sites'].append(N_site)
    SiteDb = pd.DataFrame(table)
    log_message += "\nSpecialPoint_genes {0} PID {1} SNP_counts_SiteCalc end {2}".format(scaffold, pid, time.time())

    log_message += "\nSpecialPoint_genes {0} PID {1} SNP_counts_geneCalc start {2}".format(scaffold, pid, time.time())
    table = defaultdict(list)
    for mm in sorted(xdb['mm'].unique()):
        # Filter to this mm level and set up for quick indexing
        fdb = xdb[xdb['mm'] <= mm].sort_values('mm').drop_duplicates(subset=['scaffold', 'position'], keep='last').sort_values('position').set_index('position')

        for i, row in gdb.iterrows():
            # Calculate gene length
            gLen = abs(row['end'] - row['start']) + 1

            # Subset to this gene
            db = fdb.loc[int(row['start']):int(row['end'])]

            # Report summary stuff
            table['mm'].append(mm)
            table['gene'].append(row['gene'])
            table['gene_length'].append(gLen)
            table['divergent_site_count'].append(len(db))

            # Report type counts
            for allele_count, name in zip([1, 2], ['SNS', 'SNV']):
                table['{0}_count'.format(name)].append(len(db[db['allele_count'] == allele_count]))

                for snp_type in ['N', 'S']:
                    table["{0}_{1}_count".format(name, snp_type)].append(
                    len(db[(db['allele_count'] == allele_count) & (db['mutation_type'] == snp_type)]))

    GGdb = pd.DataFrame(table).merge(SiteDb, on='gene', how='left').reset_index(drop=True)
    log_message += "\nSpecialPoint_genes {0} PID {1} SNP_counts_geneCalc end {2}".format(scaffold, pid, time.time())

    # Calculate dn/ds
    GGdb.loc[:, 'dNdS_substitutions'] = [((nC/nS) / (sC/sS)) if ((sC > 0) & (sS > 0)) else np.nan for nC, nS, sC, sS in zip(
                                GGdb['SNS_N_count'], GGdb['N_sites'],
                                GGdb['SNS_S_count'], GGdb['S_sites'])]
    GGdb.loc[:, 'pNpS_variants'] = [((nC/nS) / (sC/sS)) if ((sC > 0) & (sS > 0)) else np.nan for nC, nS, sC, sS in zip(
                                    GGdb['SNV_N_count'], GGdb['N_sites'],
                                    GGdb['SNV_S_count'], GGdb['S_sites'])]
    # GGdb['SNPs_per_bp'] = [x/y if y > 0 else np.nan for x, y in \
    #                     zip(GGdb['divergent_site_count'], GGdb['gene_length'])]

    return GGdb, log_message

def Characterize_SNPs_wrapper(Ldb, gdb, gene2sequence):
    '''
    A wrapper for characterizing SNPs

    RELIES ON HAVING gene2sequence AS A GLOBAL (needed for multiprocessing speed)

    Arguments:
        Ldb = CumulativeSNVtable for a single scaffold
        gdb = table of genes

    Returns:
        Sdb = The Cumulative SNV table with extra information added
    '''
    if len(Ldb) == 0:
        return pd.DataFrame()

    # Get a non-nonredundant list of SNPs
    Sdb = Ldb.drop_duplicates(subset=['scaffold', 'position'], keep='last')\
                .sort_index().drop(columns=['mm'])
    Sdb.loc[:, 'position'] = Sdb['position'].astype(int)

    # Filter out SNPs that shouldn't be profiled like this
    Sdb = Sdb[Sdb['cryptic'] == False]
    Sdb = Sdb.drop(columns="cryptic")

    if 'morphia' in Sdb.columns:
        col = 'morphia'
    else:
        col = 'allele_count'
    Sdb[col] = Sdb[col].astype(int)
    Sdb = Sdb[(Sdb[col] > 0) & (Sdb[col] <= 2)]

    # Make sure some SNPs to profile remain
    if len(Sdb) == 0:
        return pd.DataFrame()

    # Characterize
    sdb = characterize_SNPs(gdb, Sdb, gene2sequence)
    assert len(Sdb) == len(sdb)
    sdb = pd.merge(Sdb, sdb, on=['position'], how='left').reset_index(drop=True)

    # Return
    return sdb

def characterize_SNPs(gdb, Sdb, gene2sequence):
    '''
    Determine the type of SNP (synonymous, non-synynomous, or intergenic)

    RELIES ON HAVING gene2sequence AS A GLOBAL (needed for multiprocessing speed)
    '''
    table = defaultdict(list)
    for i, row in Sdb.iterrows():
        db = gdb[(gdb['start'] <= row['position']) & (gdb['end'] >= row['position'])]
        if len(db) == 0:
            table['position'].append(row['position'])
            table['mutation_type'].append('I')
            table['mutation'].append('')
            table['gene'].append('')
        elif len(db) > 1:
            table['position'].append(row['position'])
            table['mutation_type'].append('M')
            table['mutation'].append('')
            table['gene'].append(','.join(db['gene'].tolist()))
        else:
            # Get the original sequence
            original_sequence = gene2sequence[db ['gene'].tolist()[0]]
            if db['direction'].tolist()[0] == '-1':
                original_sequence = original_sequence.reverse_complement()

            # Make the new sequence
            snp_start = row['position'] - db['start'].tolist()[0]
            new_sequence = original_sequence.tomutable()
            new_sequence[snp_start] = row['var_base']
            if new_sequence[snp_start] == original_sequence[snp_start]:
                new_sequence[snp_start] = row['con_base']
            new_sequence = new_sequence.toseq()

            # Translate
            if db['direction'].tolist()[0] == '-1':
                old_aa_sequence = original_sequence.reverse_complement().translate()
                new_aa_sequence = new_sequence.reverse_complement().translate()
            else:
                old_aa_sequence = original_sequence.translate()
                new_aa_sequence = new_sequence.translate()

            # old_aa_sequence = original_sequence.translate()
            # new_aa_sequence = new_sequence.translate()

            # Find mutation
            mut_type = 'S'
            mut = 'S:' + str(snp_start)
            for aa in range(0, len(old_aa_sequence)):
                if new_aa_sequence[aa] != old_aa_sequence[aa]:
                    mut_type = 'N'
                    mut = 'N:' + str(old_aa_sequence[aa]) + str(snp_start) + str(new_aa_sequence[aa])
                    break

            # Store
            table['position'].append(row['position'])
            table['mutation_type'].append(mut_type)
            table['mutation'].append(mut)
            table['gene'].append(db['gene'].tolist()[0])

    return pd.DataFrame(table)

def iterate_commands(scaffolds_to_profile, Gdb, kwargs):
    '''
    Break into individual scaffolds
    '''
    processes = kwargs.get('processes', 6)
    s2g = kwargs.get('s2g', None)
    SECONDS = min(60, sum(calc_estimated_runtime(s2g[scaffold]) for scaffold in scaffolds_to_profile)/(processes+1))

    cmds = []
    seconds = 0
    for scaffold, gdb in Gdb.groupby('scaffold'):
        if scaffold not in scaffolds_to_profile:
            continue

        # make this comammand
        cmd = Command()
        cmd.scaffold = scaffold
        cmd.arguments = kwargs

        # Add estimated seconds
        seconds += calc_estimated_runtime(s2g[scaffold])
        cmds.append(cmd)

        # See if you're done
        if seconds >= SECONDS:
            yield cmds
            seconds = 0
            cmds = []

    yield cmds

def calc_estimated_runtime(pairs):
    SLOPE_CONSTANT = 0.01
    return pairs * SLOPE_CONSTANT



class Command():
    def __init__(self):
        pass

def parse_genes(gene_file_loc, **kwargs):
    '''
    Parse a file of genes based on the file extention.

    Currently supported extentions are:
        .fna (prodigal)
        .gb / .gbk (genbank)

    Methods return a table of genes (Gdb) and a dictionary of gene -> sequence
    '''
    if ((gene_file_loc[-4:] == '.fna') | (gene_file_loc[-3:] == '.fa')):
        return parse_prodigal_genes(gene_file_loc)

    elif ((gene_file_loc[-3:] == '.gb') | (gene_file_loc[-4:] == '.gbk')):
        return parse_genbank_genes(gene_file_loc)

    else:
        print("I dont know how to process {0}".format(gene_file_loc))
        raise Exception

def parse_prodigal_genes(gene_fasta):
    '''
    Parse the prodigal .fna file

    Return a datatable with gene info and a dictionary of gene -> sequence
    '''
    scaff2gene2sequence = {}
    scaff2geneinfo = {}
    for record in SeqIO.parse(gene_fasta, 'fasta'):
        gene = str(record.id)
        scaff = "_".join(gene.split("_")[:-1])

        if scaff not in scaff2geneinfo:
            scaff2geneinfo[scaff] = defaultdict(list)
        table = scaff2geneinfo[scaff]

        table['gene'].append(gene)
        table['scaffold'].append(scaff)
        table['direction'].append(record.description.split("#")[3].strip())
        table['partial'].append('partial=01' in record.description)

        # NOTE: PRODIGAL USES A 1-BASED INDEX AND WE USE 0, SO CONVERT TO 0 HERE
        table['start'].append(int(record.description.split("#")[1].strip())-1)
        table['end'].append(int(record.description.split("#")[2].strip())-1)

        if scaff not in scaff2gene2sequence:
            scaff2gene2sequence[scaff] = {}
        scaff2gene2sequence[scaff][gene] = record.seq

    scaffs = list(scaff2geneinfo.keys())
    for scaff in scaffs:
        scaff2geneinfo[scaff] = pd.DataFrame(scaff2geneinfo[scaff])

    return scaff2geneinfo, scaff2gene2sequence

def parse_genbank_genes(gene_file, gene_name='gene'):
    '''
    Parse a genbank file. Gets features marked as CDS
    '''
    scaff2geneinfo = {}
    scaff2gene2sequence = {}
    for record in SeqIO.parse(gene_file, 'gb'):
        scaffold = record.id
        for feature in record.features:
            if feature.type == 'CDS':
                gene = feature.qualifiers[gene_name][0]
                loc = feature.location
                if type(loc) is Bio.SeqFeature.CompoundLocation:
                    partial = 'compound'
                else:
                    partial = False

                if scaffold not in scaff2geneinfo:
                    scaff2geneinfo[scaffold] = defaultdict(list)
                table = scaff2geneinfo[scaffold]

                table['gene'].append(gene)
                table['scaffold'].append(scaffold)
                table['direction'].append(feature.location.strand)
                table['partial'].append(partial)

                table['start'].append(loc.start)
                table['end'].append(loc.end - 1)

                if scaffold not in scaff2gene2sequence:
                    scaff2gene2sequence[scaffold] = {}
                scaff2gene2sequence[scaffold][gene] = feature.location.extract(record).seq

    scaffs = list(scaff2geneinfo.keys())
    for scaff in scaffs:
        scaff2geneinfo[scaff] = pd.DataFrame(scaff2geneinfo[scaff])

    return scaff2geneinfo, scaff2gene2sequence

def get_gene_info(IS,  ANI_level=0):
    #IS = inStrain.SNVprofile.SNVprofile(IS_loc)

     # Get the mm level
    mm = _get_mm(IS, ANI_level)

    # Load all genes
    Gdb = IS.get('genes_table')

    # Load coverage, clonality, and SNPs
    for thing in ['genes_coverage', 'genes_clonality', 'genes_SNP_count']:
        db = IS.get(thing)
        if db is None:
            logging.debug('Skipping {0} gene calculation; you have none'.format(thing))
            continue
        if len(db) == 0:
            logging.debug('Skipping {0} gene calculation; you have none'.format(thing))
            continue
        db = db[db['mm'] <= mm].sort_values('mm').drop_duplicates(subset=['gene'], keep='last')
        del db['mm']
        Gdb = pd.merge(Gdb, db, on='gene', how='left')

    Gdb['min_ANI'] = ANI_level

    return Gdb

def _get_mm(IS, ANI):
    '''
    Get the mm corresponding to an ANI level in an IS
    '''
    if ANI > 1:
        ANI = ANI / 100

    rLen = IS.get('mapping_info')['mean_pair_length'].tolist()[0]
    mm = int(round((rLen - (rLen * ANI))))
    return mm



#!/usr/bin/env python

import copy
import gzip
import logging
import multiprocessing
import os
import random
import time
import traceback
from collections import defaultdict

import numpy as np
import pandas as pd
import pysam
from Bio import SeqIO
from numba import jit
from tqdm import tqdm

import inStrain.logUtils
import inStrain.profile.fasta
import inStrain.controller

global i2o
global v2o

i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}
v2o = {'min_read_ani':0, 'max_insert':1, 'min_insert':2, 'min_mapq':3}

class Controller():

    def main_from_profile(self, ISP, bam, **kwargs):
        '''
        The main method when called from the "profile" module

        Args:
            ISP = pre-initialized inStrain profile
            bam = location of .bam file
            args = the rest of the command line arguments

        Returns:
            Ridc = dictionary of read -> mismatches
            RR = pandas dataframe describing filtering
        '''
        detailed_report = kwargs.get('detailed_mapping_info', False)

        # Set up and parse .fasta file
        inStrain.logUtils.log_checkpoint("FilterReads", "load_fasta", "start")
        fasta_db, scaff2sequence, s2l = inStrain.profile.fasta.load_fasta(**kwargs)
        scaffolds = list(fasta_db['scaffold'].unique())
        inStrain.logUtils.log_checkpoint("FilterReads", "load_fasta", "end")

        inStrain.controller.patch_mp_connection_bpo_17560()

        # Filter the reads and store read reports
        if detailed_report:
            Rdic, RR, dRR = load_paired_reads(bam, scaffolds, **kwargs)

            # Store and delete the detailed report
            ISP.store('detailed_mapping_info', dRR, 'pandas', "Details report on reads")
            del dRR

        else:
            Rdic, RR = load_paired_reads(bam, scaffolds, **kwargs)

        # Return the Rdic and ReadReport
        return Rdic, RR, fasta_db, scaff2sequence, s2l

    def main(self, args):
        '''
        The main method when called explicitly (as its own module)
        '''
        bam = args.bam
        vargs = vars(args)
        del vargs['bam']

        detailed_report = vargs.get('detailed_mapping_info', False)
        generate_sam = vargs.get('generate_sam', False)
        out_folder = vargs.get('output', False)

        # Set up the output folder
        if not os.path.isdir(out_folder):
            os.mkdir(out_folder)

        # Set up .fasta file
        FAdb, s2s = load_fasta(args.fasta)

        # Get the paired reads
        scaffolds = list(FAdb['scaffold'].unique())
        if detailed_report:
            Rdic, RR, dRR = load_paired_reads(bam, scaffolds, **vargs)
        else:
            Rdic, RR = load_paired_reads(bam, scaffolds, **vargs)
            dRR = None

        # Make a .sam
        if generate_sam:
            print("The ability to make .sam files is not finished yet; sorry!")

        # Save results
        self.write_results(out_folder, RR, dRR, **vargs)

    def write_results(self, out_folder, RR, dRR, **kwargs):
        '''
        Save the results in a folder for the filter_reads module
        '''
        assert os.path.isdir(out_folder)

        RR_loc = os.path.join(out_folder, 'mapping_info.csv')
        write_mapping_info(RR, RR_loc, **kwargs)

        if dRR is not None:
            RR_loc = os.path.join(out_folder, 'detailed_mapping_info.csv')
            dRR.to_csv(RR_loc, index=False, sep='\t')

def read_profile_worker(read_cmd_queue, read_result_queue, bam, single_thread=False):
    '''
    Worker to filter reads
    '''
    # Apply patch
    inStrain.controller.patch_mp_connection_bpo_17560()

    # Initilize the .bam file
    bam_init = samfile = pysam.AlignmentFile(bam)

    while True:
        if not single_thread:
            cmds = read_cmd_queue.get(True)
        else:
            try:
                cmds = read_cmd_queue.get(timeout=5)
            except:
                return

        dicts, log = scaffold_profile_wrapper(cmds, bam_init)
        read_result_queue.put((dicts, log))

        # Clean up memory
        for d in dicts:
            del d
        del log
        del dicts

def load_paired_reads(bam, scaffolds, **kwargs):
    '''
    Load paired reads to be profiled

    You have this method do a lot of things because all of these things take lots of RAM, and you want them all to be cleared as soon as possible

    Return a dictionary of results. Some things that could be in it are:
        pair2infoF: A filtered dictionary of read name to number of mismatches
        RR: A summary read reaport
        RR_detailed: A detailed read report
    '''
    # Parse the kwargs
    detailed_report = kwargs.get('detailed_mapping_info', False)
    priority_reads_loc = kwargs.get('priority_reads', None)

    # Establish tallys to keep track of numbers
    tallys = {}

    # Get the pairs
    inStrain.logUtils.log_checkpoint("FilterReads", "get_paired_reads_multi", "start")
    scaff2pair2info = get_paired_reads_multi(bam, scaffolds, **kwargs)
    inStrain.logUtils.log_checkpoint("FilterReads", "get_paired_reads_multi", "end")

    if detailed_report:
        dRR = make_detailed_mapping_info(scaff2pair2info)

    # Handle paired-read filtering
    inStrain.logUtils.log_checkpoint("FilterReads", "paired_reads", "start")
    priority_reads = load_priority_reads(priority_reads_loc)
    scaff2pair2info = paired_read_filter(scaff2pair2info, priority_reads_set=priority_reads, tallys=tallys, **kwargs)
    inStrain.logUtils.log_checkpoint("FilterReads", "paired_reads", "end")

    # Filter and make the report
    inStrain.logUtils.log_checkpoint("FilterReads", "filter_reads", "start")
    scaff2pair2infoF, RR = filter_scaff2pair2info(scaff2pair2info, tallys,
                                                priority_reads_set=priority_reads,
                                                **kwargs)
    inStrain.logUtils.log_checkpoint("FilterReads", "filter_reads", "end")

    if detailed_report:
        return scaff2pair2infoF, RR, dRR
    else:
        return scaff2pair2infoF, RR

def filter_scaff2pair2info(scaff2pair2info, tallys={}, priority_reads_set=set(), **kwargs):
    '''
    Filter scaff2pair2info and generate a read report
    '''
    # Set up priority reads
    assert type(kwargs.get('priority_reads', 'none')) != type(set())
    priority_reads = priority_reads_set
    assert type(priority_reads) == type(set()), type(priority_reads)

    #item2order, to make it easier to read
    i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}

    # Calculate max insert
    max_insert_relative = kwargs.get('max_insert_relative', 3)
    median_insert = np.median([value[i2o['insert_distance']] for scaff, pair2info \
                    in scaff2pair2info.items() for pair, value in pair2info.items()\
                    if value[i2o['reads']] == 2])
    max_insert = median_insert * max_insert_relative

    # Get filter values
    values = {}
    values['min_mapq'] = kwargs.get('min_mapq', 2)
    values['max_insert'] = max_insert
    values['min_insert'] = kwargs.get('min_insert', 50)
    values['min_read_ani'] = kwargs.get('min_read_ani', 0.97)
    values['pairing_filter'] = kwargs.get('pairing_filter', 'paired_only')

    # Set up the filtered dictionary
    scaff2pair2mm = {}

    # Make tallys for individual scaffolds
    table = defaultdict(list)

    for scaff, pair2info in scaff2pair2info.items():
        # Do the tallys
        if scaff not in tallys:
            tallys[scaff] = defaultdict(int)

        # Initialize some columns
        for c in ["pass_pairing_filter", "pass_min_read_ani", "pass_max_insert",
                    "pass_min_insert", "pass_min_mapq", "filtered_pairs",
                    "filtered_singletons", "filtered_priority_reads"]:
            tallys[scaff][c] = 0

        scaff2pair2mm[scaff] = {}
        for pair, info in pair2info.items():
            update_tallys(tallys, pair, info, values, scaff, scaff2pair2mm, priority_reads)

        # Make into a table
        table['scaffold'].append(scaff)
        for key, value in tallys[scaff].items():
            table[key].append(value)


        if len(pair2info.keys()) > 0:
            # Do the means
            for i, att in enumerate(['mistmaches', 'insert_distance', 'mapq_score', 'pair_length']):
                table['mean_' + att].append(np.mean([info[i] for pair, info in pair2info.items()]))
            table['mean_PID'].append(np.mean([(1 - (float(info[i2o['nm']]) /
                                                    float(info[i2o['length']]))) for pair, info in pair2info.items()]))

            # Do a the medians
            table['median_insert'].append(np.median([info[i2o['insert_distance']] for pair, info in pair2info.items()]))
        else:
            for att in ['mistmaches', 'insert_distance', 'mapq_score', 'pair_length']:
                table['mean_' + att].append(np.nan)
            table['mean_PID'].append(np.nan)
            table['median_insert'].append(np.nan)

    try:
        Adb = pd.DataFrame(table)
    except:
        for k, v in table.items():
            print(k, len(v))
        assert False

    # Make tallys for all scaffolds
    table = defaultdict(list)
    table['scaffold'].append('all_scaffolds')

    CAdb = Adb[Adb['pass_pairing_filter'] > 0]
    total_reads = CAdb['pass_pairing_filter'].sum()

    for c in list(Adb.columns):
        if c == 'scaffold':
            pass
        elif c.startswith('mean_'):
            table[c].append(sum([v * m for v, m in zip(CAdb[c],
                                                       CAdb['pass_pairing_filter'])])/total_reads)
        elif c.startswith('median_'):
            table[c].append(sum([v * m for v, m in zip(CAdb[c],
                                                       CAdb['pass_pairing_filter'])])/total_reads)
        else:
            table[c].append(int(CAdb[c].sum()))
    adb = pd.DataFrame(table)

    # Concat
    Rdb = pd.concat([adb, Adb]).reset_index(drop=True)

    return  scaff2pair2mm, Rdb

# def update_tallys(tallys, pair, info, values, scaffold, scaff2pair2mm, priority_reads):
#     '''
#     The meat of filter_scaff2pair2info
#     '''
#     # Evaluate this pair
#     tallys[scaffold]['pass_pairing_filter'] += 1
#     f_results = evaluate_pair(info, values)
#
#     # Tally the results for what filteres passed
#     for name, index in v2o.items():
#          tallys[scaffold]['pass_' + name] += f_results[index]
#
#     # Tally the results for if the whole pair passed
#     if f_results.sum() == 4:
#         tallys[scaffold]['filtered_pairs'] += 1
#         scaff2pair2mm[scaffold][pair] = info[0]
#
#         if info[i2o['reads']] == 1:
#             tallys[scaffold]['filtered_singletons'] += 1
#
#         if pair in priority_reads:
#             tallys[scaffold]['filtered_priority_reads'] += 1
#
#
# i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}
# v2o = {'min_read_ani':0, 'max_insert':1, 'min_insert':2, 'min_mapq':3}
# def evaluate_pair(info, values):
#     '''
#     Return a list of the filters that this pair passes and fails
#     Argumnets:
#         info: np array listing info about this pair in the i2o order
#         values: dictionary listing the filters to use when evaluting this pair
#     Returns:
#         f_resilts: np array listing which filters pass (1) and fail (0) in v2o order
#     '''
#     # Initialize results for this pair
#     f_results = np.zeros(4)
#
#     # Handle PID
#     PID = 1 - (float(info[i2o['nm']]) / float(info[i2o['length']]))
#     if PID > values['min_read_ani']:
#         f_results[v2o['min_read_ani']] = 1
#
#     # Handle mapQ
#     if info[i2o['mapq']] > values['min_mapq']:
#         f_results[v2o['min_mapq']] = 1
#
#     # If this is a pair check insert distance:
#     if ((info[i2o['reads']] == 2) & (info[i2o['insert_distance']] != -1)):
#         if info[i2o['insert_distance']] > values['min_insert']:
#             f_results[v2o['min_insert']] = 1
#         if info[i2o['insert_distance']] < values['max_insert']:
#             f_results[v2o['max_insert']] = 1
#
#     # Otherwise give those a pass
#     else:
#         f_results[v2o['min_insert']] = 1
#         f_results[v2o['max_insert']] = 1
#
#     return f_results

def update_tallys(tallys, pair, info, values, scaffold, scaff2pair2mm, priority_reads):
    '''
    The meat of filter_scaff2pair2info
    '''
    # Evaluate this pair
    tallys[scaffold]['pass_pairing_filter'] += 1
    f_results = evaluate_pair(info, np.zeros(4), values['min_read_ani'], values['min_mapq'], values['min_insert'],
                              values['max_insert'])

    # Tally the results for what filteres passed
    for name, index in v2o.items():
         tallys[scaffold]['pass_' + name] += f_results[index]

    # Tally the results for if the whole pair passed
    if f_results.sum() == 4:
        tallys[scaffold]['filtered_pairs'] += 1
        scaff2pair2mm[scaffold][pair] = info[0]

        if info[i2o['reads']] == 1:
            tallys[scaffold]['filtered_singletons'] += 1

        if pair in priority_reads:
            tallys[scaffold]['filtered_priority_reads'] += 1

@jit(nopython=True)
def evaluate_pair(info, f_results, min_read_ani, min_mapq, min_insert, max_insert):
    '''
    Return a list of the filters that this pair passes and fails

    Argumnets:
        info: np array listing info about this pair in the i2o order
        values: dictionary listing the filters to use when evaluting this pair

    Returns:
        f_resilts: np array listing which filters pass (1) and fail (0) in v2o order

    i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}
    v2o = {'min_read_ani':0, 'max_insert':1, 'min_insert':2, 'min_mapq':3}
    '''
    # Initialize results for this pair
    #f_results = np.zeros(4)

    # Handle PID
    PID = 1 - (float(info[0]) / float(info[3]))
    if PID > min_read_ani:
        f_results[0] = 1

    # Handle mapQ
    if info[2] > min_mapq:
        f_results[3] = 1

    # If this is a pair check insert distance:
    if ((info[4] == 2) & (info[1] != -1)):
        if info[1] > min_insert:
            f_results[2] = 1
        if info[1] < max_insert:
            f_results[1] = 1

    # Otherwise give those a pass
    else:
        f_results[1] = 1
        f_results[2] = 1

    return f_results

def load_priority_reads(file_loc):
    '''
    Loads a file of reads and returns a set of their names
    '''
    # is it None?
    if file_loc is None:
        return set()

    # Is it zipped?
    if file_loc[-3:] == '.gz':
        reader = gzip.open(file_loc, 'rt')
    else:
        reader = open(file_loc, 'r')

    # Figure out the type
    for line in reader.readlines():
        if line[0] == '@':
            TYPE = 'fastq'
        else:
            TYPE = 'list'
        break

    reader.close()
    if file_loc[-3:] == '.gz':
        reader = gzip.open(file_loc, 'rt')
    else:
        reader = open(file_loc, 'r')

    reads = set()
    if TYPE == 'fastq':
        for line in reader.readlines():
            if line[0] != '@':
                continue
            reads.add(line[1:].strip())

    elif TYPE == 'list':
        for line in reader.readlines():
            reads.add(line.strip())

    reader.close()

    return reads

def paired_read_filter(scaff2pair2info, priority_reads_set=set(), tallys=None, **kwargs):
    '''
    Filter scaff2pair2info to keep / remove paired / unpaired reads
    '''
    assert type(kwargs.get('priority_reads', 'none')) != type(set())
    priority_reads = priority_reads_set
    pairing_filter = kwargs.get('pairing_filter', 'paired_only')
    scaff2pair2infoF = {}
    pair2scaffold = {}
    assert type(priority_reads) == type(set()), type(priority_reads)

    for scaff, p2i in scaff2pair2info.items():
        # Initilize this scaffold
        scaff2pair2infoF[scaff] = {}
        if tallys is not None:
            tallys[scaff] = defaultdict(int)
            for v in ['unfiltered_reads', 'unfiltered_pairs', 'unfiltered_singletons',
                      'unfiltered_priority_reads']:
                tallys[scaff][v] = 0

        for p, i in p2i.items():
            # Update tallys; info[4] = number of reads
            if tallys is not None:
                tallys[scaff]['unfiltered_reads'] += i[4]
                if i[4] == 2:
                    tallys[scaff]['unfiltered_pairs'] += 1
                if i[4] == 1:
                    tallys[scaff]['unfiltered_singletons'] += 1
                if p in priority_reads:
                    tallys[scaff]['unfiltered_priority_reads'] += 1

            # Determine if it's going to make it into the final set
            if pairing_filter == 'paired_only':
                if ((i[4] == 2) | (p in priority_reads)):
                    scaff2pair2infoF[scaff][p] = i

            elif pairing_filter == 'non_discordant':
                # Add it if it's not already in there
                if ((p not in pair2scaffold) | (p in priority_reads)):
                    scaff2pair2infoF[scaff][p] = i
                    pair2scaffold[p] = scaff

                # If it is already in there, that means its concordant, so delete it
                else:
                    del scaff2pair2infoF[pair2scaffold[p]][p]

            elif pairing_filter == 'all_reads':
                if p in pair2scaffold:
                    # Average the pairs
                    mi = _merge_info(i, scaff2pair2infoF[pair2scaffold[p]][p])
                    scaff2pair2infoF[scaff][p] = mi
                    scaff2pair2infoF[pair2scaffold[p]][p] = mi

                else:
                    pair2scaffold[p] = scaff
                    scaff2pair2infoF[scaff][p] = i

            else:
                logging.error("Do not know paired read filter \"{0}\"; crashing now".format(pairing_filter))
                raise Exception

    return scaff2pair2infoF

def _merge_info(i1, i2):
    #{'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}
    return np.array([i1[0] + i2[0],
            -2,
            max([i1[2] + i2[2]]),
            i1[3] + i2[3],
            i1[4] + i2[4],
            -1,
            -1], dtype="int64")

def make_detailed_mapping_info(scaff2pair2info, pairTOinfo=None, version=2):
    '''
    Make a detailed pandas dataframe from pair2info
    '''
    if pairTOinfo is None:
        pairTOinfo = dict()
    if version == 2:
        i2o = {'mm':0, 'insert_dist':1, 'mapq':2, 'length':3, 'reads':4,
                'start':5, 'stop':6}
    elif version == 1:
        i2o = {'mm':0, 'insert_dist':1, 'mapq':2, 'length':3,}

    keepers = pairTOinfo.keys()
    report_keepers =  (len(keepers) > 0)

    table = defaultdict(list)
    for scaff, pair2info in scaff2pair2info.items():
        for pair, array in pair2info.items():
            table['read_pair'].append(pair)
            table['scaffold'].append(scaff)
            if report_keepers:
                table['pass_filters'].append(pair in keepers)

            for item, location in i2o.items():
                table[item].append(array[location])

    return pd.DataFrame(table)

def load_fasta(fasta_file):
    '''
    Load the sequences to be profiled

    Return a table listing scaffold name, start, end
    '''
    # PROFILE ALL SCAFFOLDS IN THE .FASTA FILE
    scaff2sequence = SeqIO.to_dict(SeqIO.parse(fasta_file, "fasta")) # set up .fasta file
    s2l = {s:len(scaff2sequence[s]) for s in list(scaff2sequence.keys())} # Get scaffold2length
    Fdb = pd.DataFrame(list(s2l.items()), columns=['scaffold', 'end'])
    Fdb['start'] = 0

    return Fdb, scaff2sequence # also return s2l - alexcc 5/9/2019: Nah, make it scaff2sequence (s2s) (M.O. 6/10/19)

def filter_paired_reads_dict2(pair2info, **kwargs):
    '''
    Filter the dictionary of paired reads, end with read -> mm
    '''
    i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}

    # Get kwargs
    min_read_ani = kwargs.get('min_read_ani', 0.97)
    max_insert_relative = kwargs.get('max_insert_relative', 3)
    min_insert = kwargs.get('min_insert', 50)
    min_mapq = kwargs.get('min_mapq', 2)

    # Get max insert
    max_insert = np.median([value[1] for key, value in pair2info.items() if value[i2o['reads']] == 2]) * max_insert_relative

    # Return dictionary of pairs
    return {copy.deepcopy(key):copy.deepcopy(value[0])
            for key, value in pair2info.items() if _evaluate_pair2(value,
            min_read_ani=min_read_ani,
            max_insert=max_insert,
            min_insert=min_insert,
            min_mapq=min_mapq)}

def makeFilterReport2(scaff2pair2info, pairTOinfo=False, priority_reads_set=None, **kwargs):
    '''
    Make a scaffold-level report on when reads are filtered using get_paired_reads_multi2

    If you've already filtered out pairs as you'd like, pass in pairTOinfo
    '''
    if priority_reads_set is None:
        priority_reads_set = set()
    assert type(kwargs.get('priority_reads', 'none')) != type(set())
    priority_reads = priority_reads_set
    profile_scaffolds = kwargs.get('scaffold_level_mapping_info', None)

    #item2order
    i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}

    # Calculate max insert
    max_insert_relative = kwargs.get('max_insert_relative', 3)
    median_insert = np.median([value[i2o['insert_distance']] for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items() if value[i2o['reads']] == 2])
    max_insert = median_insert * max_insert_relative

    # Get values
    values = {}
    values['min_read_ani'] = kwargs.get('min_read_ani', 0.97)
    values['max_insert'] = max_insert
    values['min_insert'] = kwargs.get('min_insert', 50)
    values['min_mapq'] = kwargs.get('min_mapq', 2)

    # Make report on all scaffolds
    logging.debug('running on all reads')
    table = defaultdict(list)
    table['scaffold'].append('all_scaffolds')
    table['unfiltered_reads'].append(sum([value[i2o['reads']] for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items()]))
    table['unfiltered_pairs'].append(len([True for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items() if value[i2o['reads']] == 2]))
    table['unfiltered_singletons'].append(len([True for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items() if (value[i2o['reads']] == 1)]))
    table['unfiltered_priority_reads'].append(len([True for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items() if (pair in priority_reads)]))

    if pairTOinfo != False:
        keepers = set(pairTOinfo.keys())
        infos = [info for scaff, pair2info in scaff2pair2info.items() for pair, info in pair2info.items() if pair in keepers]
        table['pass_pairing_filter'].append(len(infos))
    else:
        infos = [info for scaff, pair2info in scaff2pair2info.items() for pair, info in pair2info.items()]

    for att, v in values.items():
        kwargs={att:v}
        table['pass_' + att].append(len([True for info in infos if (_evaluate_pair2(info, **kwargs))]))
    table['filtered_pairs'].append(len([True for info in infos if (_evaluate_pair2(info, **values))]))
    table['filtered_singletons'].append(len([True for info in infos if ((info[i2o['reads']] == 1) & (_evaluate_pair2(info, **values)))]))
    table['filtered_priority_reads'].append(len([True for scaff, pair2info in scaff2pair2info.items() for pair, info in pair2info.items() if ((pair in priority_reads) & (_evaluate_pair2(info, **values)))]))

    for i, att in enumerate(['mistmaches', 'insert_distance', 'mapq_score', 'pair_length']):
        table['mean_' + att].append(np.mean([info[i] for info in infos]))
    table['median_insert'].append(np.median([info[i2o['insert_distance']] for info in infos]))
    table['mean_PID'].append(np.mean([(1 - (float(info[i2o['nm']]) / float(info[i2o['length']]))) for info in infos]))
    Adb = pd.DataFrame(table)

    table = defaultdict(list)
    logging.debug('running on individual scaffolds')
    for scaff, pair2info in scaff2pair2info.items():
        table['scaffold'].append(scaff)

        if pairTOinfo != False:
            #keepers = set(pairTOinfo.keys()) This is calculated above; dont need twice
            infos = [info for pair, info in pair2info.items() if pair in keepers]
            table['pass_pairing_filter'].append(len(infos))
        else:
            infos = [info for pair, info in pair2info.items()]

        table['filtered_pairs'].append(len([True for info in infos if (_evaluate_pair2(info, **values))]))

        if profile_scaffolds == True:
            table['unfiltered_reads'].append(sum([value[i2o['reads']] for pair, value in pair2info.items()]))
            table['unfiltered_pairs'].append(len([True for pair, value in pair2info.items() if value[i2o['reads']] == 2]))
            table['unfiltered_singletons'].append(len([True for pair, info in pair2info.items() if (info[i2o['reads']] == 1)]))
            table['unfiltered_priority_reads'].append(len([True for pair, info in pair2info.items() if (pair in priority_reads)]))

            for att, v in values.items():
                kwargs={att:v}
                table['pass_' + att].append(len([True for info in infos if (_evaluate_pair2(info, **kwargs))]))
            table['filtered_singletons'].append(len([True for info in infos if ((info[i2o['reads']] == 1) & (_evaluate_pair2(info, **values)))]))
            table['filtered_priority_reads'].append(len([True for pair, info in pair2info.items() if ((pair in priority_reads) & (_evaluate_pair2(info, **values)))]))

            for i, att in enumerate(['mistmaches', 'insert_distance', 'mapq_score', 'pair_length']):
                table['mean_' + att].append(np.mean([info[i] for pair, info in pair2info.items()]))
            table['median_insert'].append(np.median([value[1] for key, value in pair2info.items()]))
            table['mean_PID'].append(np.mean([(1 - (float(info[i2o['nm']]) / float(info[i2o['length']]))) for pair, info in pair2info.items()]))
    Sdb = pd.DataFrame(table)

    return pd.concat([Adb, Sdb]).reset_index(drop=True)

def write_mapping_info(RR, location, **kwargs):
    # Get header materials
    values = {}
    values['min_read_ani'] = kwargs.get('min_read_ani', 0.97)
    values['max_insert_relative'] = kwargs.get('max_insert_relative', 3)
    values['min_insert'] = kwargs.get('min_insert', 50)
    values['min_mapq'] = kwargs.get('min_mapq', 2)
    values['pairing_filter'] = kwargs.get('pairing_filter', 'paired_only')

    if location is None:
        return values

    # Write header
    os.remove(location) if os.path.exists(location) else None
    f = open(location, 'a')
    f.write("# {0}\n".format(' '.join(["{0}:{1}".format(k, v) for k, v in values.items()])))

    # Write csv
    RR.to_csv(f, index=False, sep='\t')

    # Close
    f.close()

def _evaluate_pair2(value, max_insert=1000000, min_read_ani=-1, min_insert=-1,
                        min_mapq=-1):
    # calculate PID for this pair
    PID = 1 - (float(value[i2o['nm']]) / float(value[i2o['length']]))

    # If this is a pair:
    if ((value[i2o['reads']] == 2) & (value[i2o['insert_distance']] != -1)):
        # See if it passes filtering
        if ((PID > min_read_ani) & (value[i2o['insert_distance']] > min_insert) & (value[i2o['insert_distance']] < max_insert)
                & (value[i2o['mapq']] > min_mapq)):
            return True
        else:
            return False

    # If this is not a pair
    else:
        if ((PID > min_read_ani) & (value[i2o['mapq']] > min_mapq)):
            return True
        else:
            return False

def get_paired_reads_multi(bam, scaffolds, **kwargs):
    '''
    Returns scaffold 2 read 2 info

    Modified version of (get_paired_reads_multi2)2
    '''
    # Initialize dictionary
    #dicts = []
    allPair2info = {}
    p = int(kwargs.get('processes', 6))
    ret_total = kwargs.get('ret_total', False)

    # Make commands
    cmd_groups = [x for x in iterate_command_groups(scaffolds, kwargs)]

    # Lets go with spawn; see if that reduces memory usage
    ctx = multiprocessing.get_context('spawn')

    # Make queues
    read_cmd_queue = ctx.Queue()
    read_result_queue = ctx.Queue()

    for cmd_group in cmd_groups:
        read_cmd_queue.put(cmd_group)

    inStrain.logUtils.log_checkpoint("FilterReads", "multiprocessing", "start")

    # Do the multiprocessing
    if p > 1:
        logging.debug("Establishing processes")
        processes = []
        for i in range(0, p):
            processes.append(ctx.Process(target=read_profile_worker, args=(read_cmd_queue, read_result_queue, bam)))
        for proc in processes:
            proc.start()

        # Set up progress bar
        pbar = tqdm(desc='Filtering reads: ', total=len(cmd_groups))

        # Get the results
        recieved_groups = 0
        while recieved_groups < len(cmd_groups):
            result_group = read_result_queue.get()
            recieved_groups += 1
            pbar.update(1)

            assert len(result_group) == 2
            results, log = result_group
            for s, pair2info in results:
                if pair2info != False:
                    allPair2info[s] = {}
                    for k, v in pair2info.items():
                        allPair2info[s][k] = v
                #dicts.append(result)
            logging.debug(log)

        # Close multi-processing
        for proc in processes:
            proc.terminate()

        # Close progress bar
        pbar.close()

    else:
        read_profile_worker(read_cmd_queue, read_result_queue, bam, single_thread=True)
        logging.info("Done profiling genes")

        # Get the results
        recieved_groups = 0
        while recieved_groups < len(cmd_groups):
            result_group = read_result_queue.get(timeout=5)
            recieved_groups += 1

            assert len(result_group) == 2
            results, log = result_group
            for s, pair2info in results:
                if pair2info != False:
                    allPair2info[s] = {}
                    for k, v in pair2info.items():
                        allPair2info[s][k] = v
                #dicts.append(result)
            logging.debug(log)

    inStrain.logUtils.log_checkpoint("FilterReads", "multiprocessing", "end")

    return allPair2info

class read_command():
    '''
    This class just holds all of the mumbo-jumbo needed to profile a scaffold
    '''
    def __init__(self):
        pass

def iterate_read_commands(scaffolds, bam, profArgs):
    '''
    Make and iterate profiling commands
    Doing it in this way makes it use way less RAM
    '''
    for scaff in scaffolds:
        yield [scaff, bam]

def iterate_command_groups(scaffolds, profArgs):
    '''
    Break these scaffolds into a series of scaffolds

    A command group is a list of scaffolds
    '''
    cmds = []
    number_groups = int(profArgs.get('ReadGroupSize', 5000))

    if number_groups > len(scaffolds):
        for scaff in scaffolds:
            yield [scaff]

    else:
        scaffs = random.sample(scaffolds, len(scaffolds))
        for n in range(number_groups):
            yield scaffs[n::number_groups]

def scaffold_profile_wrapper(cmd_group, bam_init):
    '''
    Take a command group and get the reads
    '''
    results = []
    log = ''

    scaffolds = cmd_group
    for scaff in scaffolds:
        try:
            pair2info, cur_log = get_paired_reads(bam_init, scaff)
            results.append([scaff, pair2info])
            log += cur_log

        except Exception as e:
            print(e)
            traceback.print_exc()
            logging.error("read filtering whole scaffold exception- {0}".format(scaff))
            results.append([scaff, None])

    return results, log

def get_paired_reads(samfile, scaff):
    '''
    Filter reads from a .bam file

    Rhis gets all reads and lets you filter out pairs (or not) later.

    Returns:
        pair2info: dictionary of read pair -> (mismatches, insert distance (-1 if number of reads is not 2), mapq score (highest), combined length, number of reads,
                                                reference_position_0, reference_position_1 (only used for small things))
    '''
    log_message = inStrain.logUtils.get_worker_log('GetPairedReads', scaff, 'start')

    # item2order, to make things more readable
    i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}

    # Initialize
    pair2info = {} # Information on pairs

    try:
        iter = samfile.fetch(scaff)
    except ValueError:
        t = time.strftime('%m-%d %H:%M')
        log_message = "\n{1} DEBUG FAILURE FilterReads {0} is not in .bam file\n".format(scaff, t)
        return {}, log_message

    for read in iter:
        # Dont use unmapped reads
        if read.get_reference_positions() == []:
            continue

        # Add a read the first time you see it
        elif read.query_name not in pair2info:
            pair2info[read.query_name] = np.array([read.get_tag('NM'),
                                            -1,
                                            read.mapping_quality,
                                            read.infer_query_length(),
                                            1,
                                            read.get_reference_positions()[0],
                                            read.get_reference_positions()[-1]], dtype="int64")

        # If we've seen this read's pair before
        elif read.query_name in pair2info:
            # Adjust the number of mismatches
            pair2info[read.query_name][i2o['nm']] = int(pair2info[read.query_name][i2o['nm']]) + int(read.get_tag('NM'))

            # Adjust the number of reads
            pair2info[read.query_name][i2o['reads']] += 1

            # Adjust the combined length
            pair2info[read.query_name][i2o['length']] += read.infer_query_length()

            # Adjust the mapQ
            pair2info[read.query_name][i2o['mapq']] = max(pair2info[read.query_name][i2o['mapq']], read.mapping_quality)

            # If the number of reads is 2, calculate insert size
            if pair2info[read.query_name][i2o['reads']] == 2:
                if read.get_reference_positions()[-1] > pair2info[read.query_name][i2o['start']]:
                    pair2info[read.query_name][i2o['insert_distance']] = read.get_reference_positions()[-1] - pair2info[read.query_name][i2o['start']]
                else:
                    pair2info[read.query_name][i2o['insert_distance']] = pair2info[read.query_name][i2o['stop']] - read.get_reference_positions()[0]

            # Otherwise get rid of insert distance
            else:
                pair2info[read.query_name][i2o['insert_distance']] = -1

            # Get rid of start and stop; not needed anymore
            pair2info[read.query_name][i2o['start']] = 0
            pair2info[read.query_name][i2o['stop']] = 0

    log_message += inStrain.logUtils.get_worker_log('GetPairedReads', scaff, 'end')

    return pair2info, log_message


#!/usr/bin/env python

"""
None of this code is in use at the moment
"""

def greedy_main(RCprof, names, Sprofiles, scaffolds_to_compare, s2l, **kwargs):
    '''
    DEPRECATED

    Perform greedy clustering instead of all-vs-all comparisons
    '''
    g_ani = kwargs.get('g_ani', 0.99)
    g_cov = kwargs.get('g_cov', 0.5)
    g_mm = kwargs.get('g_mm', 100)

    # Set up the clustering
    name2cluster = {}
    cluster2rep = {}
    name2Sprofile = {n:s for n, s in zip(names, Sprofiles)}

    # Set up the initial cluster
    clusters = 0
    name2cluster[names[0]] = clusters
    cluster2rep[clusters] = names[0]
    clusters += 1

    # Make sure the names are unique
    assert len(names) == len(set(names)), "IS objects do not have unique names!"

    # Figure out the total bin length
    if kwargs.get('scaffolds', None) != None:
        scaffolds = inStrain.profile.fasta.load_scaff_list(kwargs.get('scaffolds', None))
        BIN_LENGTH = sum([s2l[s] for s in scaffolds])
    else:
        BIN_LENGTH = sum([l for s, l in s2l.items()])
    logging.info("BIN_LENGTH is {0}".format(BIN_LENGTH))

    # Do the iterations
    ndbs = []
    ddbs = []
    for name, IS in zip(names[1:], Sprofiles[1:]):

        # Get a list of Sprofiles to compare to
        To_compare = []
        compare_names = []
        for i in range(0, clusters):
            To_compare.append(name2Sprofile[cluster2rep[i]])
            compare_names.append(cluster2rep[i])

        # Get the distance matrix
        Ddb, Ndb = compare_Sprofiles_wrapper(IS, To_compare, name, compare_names, scaffolds_to_compare, s2l, BIN_LENGTH, **kwargs)
        ndbs.append(Ndb)
        ddbs.append(Ddb)

        # Adjust the clusters
        found = False
        for i, row in Ddb.iterrows():
            if (row['popANI'] >= g_ani) & (row['cov'] >= g_cov):
                found = True
                logging.debug("{0} is in the same cluster as {1}".format(name, row['name2']))
                name2cluster[name] = name2cluster[row['name2']]
                break

        if not found:
            logging.debug("{0} is a new cluster".format(name))
            name2cluster[name] = clusters
            cluster2rep[clusters] = name
            clusters += 1

    # Make the output
    Cdb = pd.DataFrame(list(name2cluster.items()), columns=['name', 'cluster'])
    Ndb = pd.concat(ndbs)
    Ddb = pd.concat(ddbs)

    # Store results
    outbase = RCprof.get_location('output') + os.path.basename(RCprof.get('location')) + '_'

    RCprof.store('greedy_clusters', Cdb, 'pandas', 'Cluster affiliations of the requested IS objects')
    RCprof.store('scaffold2length', s2l, 'dictionary', 'Scaffold to length')
    RCprof.store('comparisonsTable_greedy', Ndb, 'pandas', 'Comparisons between the requested IS objects done from a greedy clustering')
    RCprof.store('parsed_comparisonsTable_greedy', Ddb, 'pandas', 'Parsed comparisons between the requested IS objects done from a greedy clustering')

    Cdb.to_csv(outbase + 'greedyClusters.tsv', index=False, sep='\t')
    Ddb.to_csv(outbase + 'parsed_comparisonsTable_greedy.tsv', index=False, sep='\t')

def compare_Sprofiles_wrapper(IS1, IS_list, name1, names, scaffolds_to_compare,
                                s2l, BIN_LENGTH, **kwargs):
    '''
    Compare IS1 to every IS in the IS_list
    '''
    table = defaultdict(list)
    cdbs = []
    for cIS, name2 in zip(IS_list, names):
        results = compare_Sprofiles(IS1, cIS, [name1, name2], scaffolds_to_compare, s2l, BIN_LENGTH, **kwargs)
        Cdb, ANI, cov = results
        table['name1'].append(name1)
        table['name2'].append(name2)
        table['ANI'].append(ANI)
        table['cov'].append(cov)
        for thing in ['g_ani', 'g_cov', 'g_mm']:
            table[thing].append(kwargs.get(thing, np.nan))
        cdbs.append(Cdb)

    Ndb = pd.concat(cdbs)
    Ddb = pd.DataFrame(table)
    return Ddb, Ndb

def compare_Sprofiles(IS1, cIS, names, scaffolds_to_compare, s2l, BIN_LENGTH, **kwargs):
    '''
    Compare all scaffolds of two Sprofiles
    '''
    Cdb, scaff2pair2mm2SNPs, scaff2pair2mm2cov = compare_scaffolds(names, [IS1, cIS], scaffolds_to_compare, s2l, **kwargs)
    ANI, cov = calc_cov_ani(Cdb, BIN_LENGTH, **kwargs)
    return Cdb, ANI, cov

def calc_cov_ani(Cdb, BIN_LENGTH, **kwargs):
    '''
    Return ANI and coverage assuming all scaffolds belong to the same genome
    '''
    g_ani = kwargs.get('g_ani', 0.99)
    g_cov = kwargs.get('g_cov', 0.5)
    g_mm = kwargs.get('g_mm', 100)

    db = Cdb[Cdb['mm'] <= g_mm].sort_values('mm').drop_duplicates(subset=['scaffold', 'name1', 'name2'], keep='last')
    tcb = sum(db['compared_bases_count'])

    if tcb == 0:
        popANI = np.nan
    else:
        popANI = sum(a * c  if a == a else 0 for a, c in
                            zip(db['popANI'], db['compared_bases_count'])) / tcb

    return popANI, (tcb / BIN_LENGTH)